@@ -1,19 +1,24 @@
-/* Zgv v2.7 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
- * Copyright (C) 1993-1995 Russell Marks. See README for license details.
+/* Zgv v3.0 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993-1998 Russell Marks. See README for license details.
  *
- * font.c - provides a font of sorts for use via svgalib
+ * font.c - provides a font of sorts for use via svgalib, and
+ *		code to use (nicer) BDF fonts converted with bdf2h.
  */
 
 
-/* it's all hard-coded and a bit ugly, so you probably won't want
- * much to look at it (hint) :)
+/* all the line-based font stuff is hard-coded and a bit ugly, so you
+ * probably won't want much to look at that (hint) :-)
  */
 
  
 #include <stdio.h>
 #include <string.h>
 #include <vga.h>
+#include "zgv.h"
 #include "font.h"
+#include "fontbmps.h"
+#include "rc_config.h"
+#include "rcfile.h"
 
 #define TABSIZE  64   /* size of a tab, in pixels */
 
@@ -24,6 +29,17 @@ static int go_through_the_motions=0;
 static int stop_after_this_x=NO_CLIP_FONT;
 
 
+/* the copyright/license for the bitmap fonts says that all
+ * copies of the fonts must contain the copyright. Since a copy
+ * was embedded in the program above (`fontbmps.h'), we'd
+ * best have this:
+ */
+static char *bitmap_fonts_copyright=
+"The bitmap fonts zgv uses are: \
+Copyright 1984-1989, 1994 Adobe Systems Incorporated. \
+Copyright 1988, 1994 Digital Equipment Corporation.";
+
+
 /* prototypes */
 int vgadrawtext(int x, int y, int siz, char *str);
 void fontcircle(int x, int y, int r);
@@ -37,6 +53,16 @@ void fontc_ll(int x, int y, int r);
 void fontc_lr(int x, int y, int r);
 int vgatextsize(int sizearg, char *str);
 void set_max_text_width(int width);
+int vgadrawtext_bmp(int xpos,int ypos,int siz,char *str);
+
+
+/* unused dummy routine to stop -Wall complaining about the copyright
+ * message not being used.
+ */
+static void its_too_damn_picky_sometimes()
+{
+printf(bitmap_fonts_copyright);
+}
 
 
 
@@ -45,6 +71,8 @@ int vgadrawtext(int x,int y,int siz,char *str)
 {
 int f,a,b,c,s1,s2,s3,s4,s5,s6,gap;
 
+if(!cfg.linetext) return(vgadrawtext_bmp(x,y,siz,str));
+
 b=y;
 a=x;
 s1=siz; s2=s1<<1; s3=s1+s2; s4=s2<<1; s5=s4+s1; s6=s3<<1;
@@ -59,11 +87,11 @@ for(f=0;f<strlen(str);f++)
     {
     int tmp;
     
-    /* print an ellipsis... well, three dots :) */
+    /* print an ellipsis... well, three dots :-) */
     /* blast the width restriction to stop possible infinite recursion */
     tmp=stop_after_this_x;
     set_max_text_width(NO_CLIP_FONT);
-    vgadrawtext(a,y,siz,"...");
+    a+=vgadrawtext(a,y,siz,"...");
     stop_after_this_x=tmp;
     /* now give up */
     break;
@@ -73,25 +101,25 @@ for(f=0;f<strlen(str);f++)
   /*** 1st step: cover some common occurances ***/
   if(!go_through_the_motions)   /* only draw it if we really want to */
   
-  if(index("abdgopq68",c)!=NULL)   /* common circle position */
+  if(strchr("abdgopq68",c)!=NULL)   /* common circle position */
     fontcircle(a+s1,b+s3,siz);
   else
     {    /* common part-circle positions */
-    if(index("cehmnrs",c)!=NULL)
+    if(strchr("cehmnrs",c)!=NULL)
       fontc_ul(a+s1,b+s3,siz);
-    if(index("ehmnrBS35",c)!=NULL)
+    if(strchr("ehmnrBS35",c)!=NULL)
       fontc_ur(a+s1,b+s3,siz);
-    if(index("cetuyCGJOQSUl035",c)!=NULL)
+    if(strchr("cetuyCGJOQSUl035",c)!=NULL)
       fontc_ll(a+s1,b+s3,siz);
-    if(index("suyBCDGJOQSU035",c)!=NULL)
+    if(strchr("suyBCDGJOQSU035",c)!=NULL)
       fontc_lr(a+s1,b+s3,siz);
     /* common line */
-    if(index("BDEFHKLMNPR",c)!=NULL)
+    if(strchr("BDEFHKLMNPR",c)!=NULL)
       vga_drawline(a,b,a,b+s4);
     }
 
 
-  /*** 2nd step: fill in rest - this is the *really* long, messy bit :) ***/
+  /*** 2nd step: fill in rest - this is the *really* long, messy bit :-) ***/
   
   /*** 2a: lowercase letters ***/
   if(!go_through_the_motions)
@@ -304,15 +332,15 @@ for(f=0;f<strlen(str);f++)
     
   /*** 3rd part: finally, move along for the next letter ***/
   /*** we do this even if go_through_the_motions is set */
-  if(index("ltfijk-. ()",c)!=NULL)
+  if(strchr("ltfijk-. ()",c)!=NULL)
     a+=s1;
   else
     {
-    if(index("?/%abcdeghnopqrsuvxyzABCDEFGHIJKLNOPQRSTUVXYZ0123456789",c)!=NULL)
+    if(strchr("?/%abcdeghnopqrsuvxyzABCDEFGHIJKLNOPQRSTUVXYZ0123456789",c)!=NULL)
       a+=s2;
     else
       {
-      if(index("mwMW",c)!=NULL)
+      if(strchr("mwMW",c)!=NULL)
         a+=s4;
       else
         {
@@ -427,3 +455,128 @@ void set_max_text_width(int width)
 {
 stop_after_this_x=width;
 }
+
+
+
+int vgadrawtext_bmp(int xpos,int ypos,int siz,char *str)
+{
+static unsigned char scanbit[128];
+static int old_colour=-1;
+int f,c,x,y;
+int initxpos=xpos;
+signed char *ptr,*fontdat;
+int ox,oy,w,h,dw;
+int yst;
+int xrunst=-1;
+int *fonttbl;
+int fontyofs,fontfh,fontoy;
+int widthadd=0;
+int m_ellipsis_size;
+
+if(current_colour!=old_colour)
+  memset(scanbit,old_colour=current_colour,sizeof(scanbit));
+
+/* pick font to match siz.
+ * It would be nicer if these were in a struct or something...! (XXX)
+ * (That's really a problem with bdf2h.c rather than this, though.)
+ */
+switch(siz)
+  {
+  case 5:
+    fontdat=font3dat; fonttbl=font3tbl;
+    fontyofs=font3yofs; fontfh=font3fh; fontoy=font3oy;
+    break;
+
+  case 4:
+    widthadd=1;
+    /* falls through */
+  
+  case 3:
+    fontdat=font2dat; fonttbl=font2tbl;
+    fontyofs=font2yofs; fontfh=font2fh; fontoy=font2oy;
+    break;
+    
+  /* this doesn't appear normally, but will if someone hacks
+   * on zgv blindly. It gently gives them a clue. :-)
+   */
+  default:
+    fprintf(stderr,"unsupported text size - treating siz as 2\n");
+    /* falls through */
+  
+  case 2:
+    fontdat=font1dat; fonttbl=font1tbl;
+    fontyofs=font1yofs; fontfh=font1fh; fontoy=font1oy;
+    break;
+  }
+
+/* m_ellipsis_size is the size that an ellipsis will take up, plus the
+ * widest possible letter (which we assume is M).
+ */
+m_ellipsis_size=3*(widthadd+fontdat[fonttbl['.'-32]+4])+
+		   widthadd+fontdat[fonttbl['M'-32]+4];
+
+for(f=0;f<strlen(str);f++)
+  {
+  if(xpos-initxpos>stop_after_this_x-m_ellipsis_size)
+    {
+    int tmp;
+    
+    /* print an ellipsis... well, three dots :-) */
+    /* blast the width restriction to stop possible infinite recursion */
+    tmp=stop_after_this_x;
+    set_max_text_width(NO_CLIP_FONT);
+    xpos+=vgadrawtext_bmp(xpos,ypos,siz,"...");
+    stop_after_this_x=tmp;
+    /* now give up */
+    break;
+    }
+  c=str[f];
+  if(c=='\t')
+    {
+    xpos=((xpos/TABSIZE)+1)*TABSIZE;
+    continue;
+    }
+  if(c<32 || c>126) c='_';
+  ptr=fontdat+fonttbl[c-32];
+  ox=*ptr++; oy=*ptr++;
+  w =*ptr++; h =*ptr++;
+  dw=*ptr++;
+  yst=fontfh-(oy-fontoy)-fontyofs-h;
+  
+  if(!go_through_the_motions && c!='\t')
+    for(y=ypos+yst;y<ypos+yst+h;y++)
+      {
+      for(x=xpos+ox;x<xpos+ox+w;x++)
+#if 0
+        /* this is the obvious way... */
+        if(*ptr++) vga_drawpixel(x,y);
+#else
+        /* ...but this is a little faster, I think, especially
+         * for large fonts. (It uses horizontal lines where possible.)
+         */
+        if(*ptr++)
+          {if(xrunst==-1) xrunst=x;}
+        else
+          if(xrunst!=-1)
+            {
+            if(xrunst==x-1)
+              vga_drawpixel(xrunst,y);
+            else
+              vga_drawscansegment(scanbit,xrunst,y,x-xrunst);
+            xrunst=-1;
+            }
+      
+      if(xrunst!=-1)
+        if(xrunst==x-1)
+          vga_drawpixel(xrunst,y);
+        else
+          vga_drawscansegment(scanbit,xrunst,y,x-xrunst);
+      xrunst=-1;
+#endif
+      }
+  
+  xpos+=dw+widthadd;
+  }
+
+return(xpos-initxpos);
+}
