@@ -8,7 +8,6 @@
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
-#include <malloc.h>
 #include <stdlib.h>
 #include "zgv.h"
 #include "gifeng.h"
@@ -186,7 +185,7 @@ void getgifinfo(PICINFO *ginfo)
 {
 ginfo->width=width;
 ginfo->height=height;
-gifhed.wide_lo=0;  /* yes, it's nasty, but wide_lo is unimportant anyway :) */
+gifhed.wide_lo=0;  /* yes, it's nasty, but wide_lo is unimportant anyway :-) */
 strcpy(ginfo->type,gifhed.sig+3);
 ginfo->bpp=bpp;
 ginfo->numcols=numcols;
