@@ -7,6 +7,8 @@
 
 #include "mouse.h"
 
+struct cfgrgb_tag { int r,g,b; };
+
 struct zgv_config {
   int videomode,zoom,vkludge,jpeg24bit,jpegspeed;
   int jpegindexstyle;  /* 0=v. fast, 1=fast, 2=normal, 3=slow (as in <=v2.7) */
@@ -20,16 +22,17 @@ struct zgv_config {
   int repeat_timer,tag_timer;
   int revert,gamma,selecting;
   int fs16col;
+  int viewer16col;
   int fullsel;
   int echotagged;
+  int forgetoldpos;
+  int linetext;
+  int smallfiletext;
   
   int brightness;
   double contrast;
-  int black_r,black_g,black_b;
-  int dark_r,dark_g,dark_b;
-  int medium_r,medium_g,medium_b;
-  int light_r,light_g,light_b;
-  int marked_r,marked_g,marked_b;
+  
+  struct cfgrgb_tag black,dark,medium,light,marked;
   int mode_allowed[256];
   int showxvpicdir;
 
