@@ -12,30 +12,97 @@
 #include "rc_config.h"
 #include "zgv.h"	/* for ZGV_VER */
 
-#define boolvalue(x) (((!strncmp(x,"on",2)) || x[0]=='y')?1:0)
 
-struct zgv_config cfg;
+#ifndef GLASTMODE
+/* not sure if this is possible, but just in case...
+ * (this is fairly conservative, of course)
+ */
+#define GLASTMODE	G1280x1024x16M
+#endif
 
 
 /* prototypes */
 void getconfig(void);
 void parseconfig(FILE *in);
 int getmodenumber(char *tptr, int *mp);
-void getrgbval(char *tptr, int *rp, int *gp, int *bp);
-void getbool(char *tptr, int *bp);
-void getint(char *tptr, int *ip);
-void getdouble(char *tptr, double *dp);
+void getrgbval(int line,void *arg2,void *arg3);
+void getbool(int line,void *arg2,void *arg3);
+void getint(int line,void *arg2,void *arg3);
+void getdouble(int line,void *arg2,void *arg3);
+mouse_type strmousetype(char *arg);
+void getmousearg(int line,void *arg2,void *arg3);
+void getjis(int line,void *arg2,void *arg3);
+void do_allmodesbad(int line,void *arg2,void *arg3);
+void do_allmodesgood(int line,void *arg2,void *arg3);
+void do_badmode(int line,void *arg1,void *arg3);
+void do_goodmode(int line,void *arg1,void *arg3);
+void do_startmode(int line,void *arg1,void *arg3);
 void defaultcfg(void);
 void findtoken(char **ptr);
 int tokenlength(char *ptr);
 int tokencompare(char *tptr, char *txt);
 int modematch(int x, int y, int bpp);
-int parsecommandline(int argc, char **argv);
+int parsecommandline(int argc, char *argv[]);
 void usage_help(void);
-void getmousearg(char *tptr,mouse_type *mp);
 
  
 
+struct zgv_config cfg;
+
+struct cfglookup_tag
+  {
+  char optname[32];
+  void (*routine)(int,void *,void *);
+  void *arg3;	/* arg1 is line no., arg2 is the after-token-pointer */
+  };
+
+struct cfglookup_tag cfglookup[]=
+  {
+  {"allmodesbad",	do_allmodesbad,	NULL},
+  {"allmodesgood",	do_allmodesgood,NULL},
+  {"badmode",		do_badmode,	NULL},
+  {"betterpgm",		getbool,	&cfg.betterpgm},
+  {"black",		getrgbval,	&cfg.black},
+  {"blockcursor",	getbool,	&cfg.blockcursor},
+  {"brightness",	getint,		&cfg.brightness},
+  {"center",		getbool,	&cfg.centreflag},
+  {"centre",		getbool,	&cfg.centreflag},
+  {"cleartext",		getbool,	&cfg.cleartext},
+  {"contrast",		getdouble,	&cfg.contrast},
+  {"dark",		getrgbval,	&cfg.dark},
+  {"force16fs",		getbool,	&cfg.force16fs},
+  {"forgetoldpos",	getbool,	&cfg.forgetoldpos},
+  {"fs16col",		getbool,	&cfg.fs16col},
+  {"fullsel",		getbool,	&cfg.fullsel},
+  {"gamma",		getbool,	&cfg.gamma},
+  {"goodmode",		do_goodmode,	NULL},
+  {"hicolmodes",	getbool,	&cfg.hicolmodes},
+  {"jpeg24bit",		getbool,	&cfg.jpeg24bit},
+  {"jpegindexstyle",	getjis,		&cfg.jpegindexstyle},
+  {"jpegspeed",		getint,		&cfg.jpegspeed},
+  {"light",		getrgbval,	&cfg.light},
+  {"linetext",		getbool,	&cfg.linetext},
+  {"medium",		getrgbval,	&cfg.medium},
+  {"mouse",		getmousearg,	&cfg.mtype},
+  {"nodelprompt",	getbool,	&cfg.nodelprompt},
+  {"onefile_progress",	getbool,	&cfg.onefile_progress},
+  {"perfectindex",	getbool,	&cfg.perfectindex},
+  {"revert",		getbool,	&cfg.revert},
+  {"showxvpicdir",	getbool,	&cfg.showxvpicdir},
+  {"smallfiletext",	getbool,	&cfg.smallfiletext},
+  {"startmode",		do_startmode,	NULL},
+  {"tagged",		getrgbval,	&cfg.marked},
+  {"tagtimeout",	getint,		&cfg.tag_timer},
+  {"thicktext",		getbool,	&cfg.thicktext},
+  {"viewer16col",	getbool,	&cfg.viewer16col},
+  {"visual",		getbool,	&cfg.xvpic_index},
+  {"vkludge",		getbool,	&cfg.vkludge},
+  {"zoom",		getbool,	&cfg.zoom},
+  {"",NULL,NULL}
+  };
+
+
+
 void getconfig()
 {
 FILE *in;
@@ -69,145 +136,31 @@ if(cfg.jpegspeed>3) cfg.jpegspeed=3;
 void parseconfig(in)
 FILE *in;
 {
-char inpline[1024],*ptr;
-int ln,f;
+static char inpline[1024];
+char *ptr;
+int ln,f,c,inpc,found=0;
 
 ln=0;
-while(fgets(inpline,1024,in)!=NULL) 
+while(fgets(inpline,sizeof(inpline),in)!=NULL) 
   {
   ln++;
   if(inpline[strlen(inpline)-1]=='\n') inpline[strlen(inpline)-1]=0;
   if((ptr=strchr(inpline,'#'))!=NULL)
     *ptr=0;
-  if(strlen(inpline)>0)
+  if(inpline[0])
     {
-    /* XXX this is begging to be converted into a lookup table or something */
+    found=0;
     ptr=inpline;
     findtoken(&ptr);
-    if(tokencompare(ptr,"zoom"))
-      getbool(ptr,&cfg.zoom);
-    else
-    if(tokencompare(ptr,"fullsel"))
-      getbool(ptr,&cfg.fullsel);
-    else
-    if(tokencompare(ptr,"revert"))
-      getbool(ptr,&cfg.revert);
-    else
-    if(tokencompare(ptr,"gamma"))
-      getbool(ptr,&cfg.gamma);
-    else
-    if(tokencompare(ptr,"force16fs"))
-      getbool(ptr,&cfg.force16fs);
-    else
-    if(tokencompare(ptr,"fs16col"))
-      getbool(ptr,&cfg.fs16col);
-    else
-    if(tokencompare(ptr,"blockcursor"))
-      getbool(ptr,&cfg.blockcursor);
-    else
-    if(tokencompare(ptr,"thicktext"))
-      getbool(ptr,&cfg.thicktext);
-    else
-    if(tokencompare(ptr,"hicolmodes"))
-      getbool(ptr,&cfg.hicolmodes);
-    else
-    if(tokencompare(ptr,"nodelprompt"))
-      getbool(ptr,&cfg.nodelprompt);
-    else
-    if(tokencompare(ptr,"onefile_progress"))
-      getbool(ptr,&cfg.onefile_progress);
-    else
-    if(tokencompare(ptr,"cleartext"))
-      getbool(ptr,&cfg.cleartext);
-    else
-    if(tokencompare(ptr,"perfectindex"))
-      getbool(ptr,&cfg.perfectindex);
-    else
-    if(tokencompare(ptr,"visual"))
-      getbool(ptr,&cfg.xvpic_index);
-    else
-    if(tokencompare(ptr,"betterpgm"))
-      getbool(ptr,&cfg.betterpgm);
-    else
-    if(tokencompare(ptr,"centre") || tokencompare(ptr,"center"))
-      getbool(ptr,&cfg.centreflag);
-    else
-    if(tokencompare(ptr,"jpeg24bit"))
-      getbool(ptr,&cfg.jpeg24bit);
-    else
-    if(tokencompare(ptr,"mouse"))
-      getmousearg(ptr,&cfg.mtype);
-    else
-    if(tokencompare(ptr,"jpegspeed"))
-      getint(ptr,&cfg.jpegspeed);
-    else
-    if(tokencompare(ptr,"vkludge"))
-      getbool(ptr,&cfg.vkludge);
-    else
-    if(tokencompare(ptr,"black"))
-      getrgbval(ptr,&cfg.black_r,&cfg.black_g,&cfg.black_b);
-    else
-    if(tokencompare(ptr,"dark"))
-      getrgbval(ptr,&cfg.dark_r,&cfg.dark_g,&cfg.dark_b);
-    else
-    if(tokencompare(ptr,"medium"))
-      getrgbval(ptr,&cfg.medium_r,&cfg.medium_g,&cfg.medium_b);
-    else
-    if(tokencompare(ptr,"light"))
-      getrgbval(ptr,&cfg.light_r,&cfg.light_g,&cfg.light_b);
-    else
-    if(tokencompare(ptr,"tagged"))
-      getrgbval(ptr,&cfg.marked_r,&cfg.marked_g,&cfg.marked_b);
-    else
-    if(tokencompare(ptr,"contrast"))
-      getdouble(ptr,&cfg.contrast);
-    else
-    if(tokencompare(ptr,"brightness"))
-      getint(ptr,&cfg.brightness);
-    else
-    if(tokencompare(ptr,"tagtimeout"))
-      getint(ptr,&cfg.tag_timer);
-    else
-    if(tokencompare(ptr,"showxvpicdir"))
-      getbool(ptr,&cfg.showxvpicdir);
-    else
-    if(tokencompare(ptr,"jpegindexstyle"))
-      {
-      getint(ptr,&cfg.jpegindexstyle);
-      if(cfg.jpegindexstyle<1) cfg.jpegindexstyle=1;
-      if(cfg.jpegindexstyle>3) cfg.jpegindexstyle=3;
-      }
-    else
-    if(tokencompare(ptr,"allmodesbad"))
-      for(f=0;f<256;f++) cfg.mode_allowed[f]=0;
-    else
-    if(tokencompare(ptr,"allmodesgood"))
-      {
-      /* we still disallow ones it hasn't got */
-      for(f=0;f<31;f++)
-        cfg.mode_allowed[f]=vga_hasmode(f)?1:0;
-      }
-    else
-    if((tokencompare(ptr,"badmode"))||(tokencompare(ptr,"goodmode")))
-      {
-      if(getmodenumber(ptr,&f)==-1)
+    inpc=*ptr;
+    for(f=0;(c=cfglookup[f].optname[0]);f++)
+      if(inpc==c && tokencompare(ptr,cfglookup[f].optname))
         {
-        fprintf(stderr,"Mode not found on line #%d of rc file.\n",ln);
-        exit(1);
+        (*cfglookup[f].routine)(ln,ptr,cfglookup[f].arg3);
+        found=1;
+        break;
         }
-      cfg.mode_allowed[f]=tokencompare(ptr,"goodmode");
-      }
-    else
-    if(tokencompare(ptr,"startmode"))
-      {
-      if(getmodenumber(ptr,&f)==-1)
-        {
-        fprintf(stderr,"Mode not found on line #%d of rc file.\n",ln);
-        exit(1);
-        }
-      cfg.videomode=f;
-      }
-    else
+    if(!found)
       {
       fprintf(stderr,"Error in line #%d of rc file.\n",ln);
       exit(1);
@@ -249,25 +202,31 @@ else
 }
 
 
-void getrgbval(char *tptr,int *rp,int *gp,int *bp)
+void getrgbval(int line,void *arg2,void *arg3)
 {
+char *tptr=(char *)arg2;
+struct cfgrgb_tag *p=(struct cfgrgb_tag *)arg3;
+
 /* skip past the current token */
 tptr+=tokenlength(tptr);
 findtoken(&tptr);
-*rp=atoi(tptr);
+p->r=atoi(tptr);
 tptr+=tokenlength(tptr);
 
 findtoken(&tptr);
-*gp=atoi(tptr);
+p->g=atoi(tptr);
 tptr+=tokenlength(tptr);
 
 findtoken(&tptr);
-*bp=atoi(tptr);
+p->b=atoi(tptr);
 }
 
 
-void getbool(char *tptr,int *bp)
+void getbool(int line,void *arg2,void *arg3)
 {
+char *tptr=(char *)arg2;
+int *bp=(int *)arg3;
+
 /* skip past the current token */
 tptr+=tokenlength(tptr);
 findtoken(&tptr);
@@ -278,8 +237,11 @@ else
 }
 
 
-void getint(char *tptr,int *ip)
+void getint(int line,void *arg2,void *arg3)
 {
+char *tptr=(char *)arg2;
+int *ip=(int *)arg3;
+
 /* skip past the current token */
 tptr+=tokenlength(tptr);
 findtoken(&tptr);
@@ -287,8 +249,11 @@ findtoken(&tptr);
 }
 
 
-void getdouble(char *tptr,double *dp)
+void getdouble(int line,void *arg2,void *arg3)
 {
+char *tptr=(char *)arg2;
+double *dp=(double *)arg3;
+
 /* skip past the current token */
 tptr+=tokenlength(tptr);
 findtoken(&tptr);
@@ -311,8 +276,11 @@ return(NO_MOUSE);
 }
 
 
-void getmousearg(char *tptr,mouse_type *mp)
+void getmousearg(int line,void *arg2,void *arg3)
 {
+char *tptr=(char *)arg2;
+mouse_type *mp=(mouse_type *)arg3;
+
 /* skip past the current token */
 tptr+=tokenlength(tptr);
 findtoken(&tptr);
@@ -320,6 +288,75 @@ findtoken(&tptr);
 }
 
 
+void getjis(int line,void *arg2,void *arg3)
+{
+int *jisp=(int *)arg3;
+getint(line,arg2,arg3);
+if(*jisp<1) *jisp=1;
+if(*jisp>3) *jisp=3;
+}
+
+
+void do_allmodesbad(int line,void *arg2,void *arg3)
+{
+int f;
+
+for(f=0;f<256;f++) cfg.mode_allowed[f]=0;
+}
+
+
+void do_allmodesgood(int line,void *arg2,void *arg3)
+{
+int f;
+
+/* we still disallow ones it hasn't got */
+for(f=0;f<=GLASTMODE;f++)
+  cfg.mode_allowed[f]=vga_hasmode(f)?1:0;
+}
+
+
+void do_badmode(int line,void *arg1,void *arg3)
+{
+char *ptr=(char *)arg1;
+int f;
+
+if(getmodenumber(ptr,&f)==-1)
+  {
+  fprintf(stderr,"Mode not found on line #%d of rc file.\n",line);
+  exit(1);
+  }
+cfg.mode_allowed[f]=0;
+}
+
+
+void do_goodmode(int line,void *arg1,void *arg3)
+{
+char *ptr=(char *)arg1;
+int f;
+
+if(getmodenumber(ptr,&f)==-1)
+  {
+  fprintf(stderr,"Mode not found on line #%d of rc file.\n",line);
+  exit(1);
+  }
+cfg.mode_allowed[f]=1;
+}
+
+
+void do_startmode(int line,void *arg1,void *arg3)
+{
+char *ptr=(char *)arg1;
+int f;
+
+if(getmodenumber(ptr,&f)==-1)
+  {
+  fprintf(stderr,"Mode not found on line #%d of rc file.\n",line);
+  exit(1);
+  }
+cfg.videomode=f;
+}
+
+
 void defaultcfg()
 {
 int f;
@@ -352,18 +389,26 @@ cfg.revert=1;		/* auto-reset scaling to off between pics */
 cfg.gamma=1;		/* try to fake more greys/colours in 8-bit modes */
 cfg.selecting=0;	/* no selection normally */
 cfg.fs16col=0;		/* normally use mono in 16-col file selector */
+cfg.viewer16col=0;	/* normally use mono in 16-col viewer mode */
 cfg.echotagged=0;	/* echo tagged files on exit */
+cfg.forgetoldpos=0;	/* don't return to old position in revisited dir */
+cfg.linetext=0;		/* use old line-based text instead of bitmap fonts */
+/* but bitmap fonts are far too slow in 16-colour mode, so stick
+ * with the old line-based font for that by default:
+ */
+if(!vga_hasmode(G640x480x256)) cfg.linetext=1;
+cfg.smallfiletext=0;	/* use smaller font for filenames in file selector */
 cfg.contrast=(double)1;
-cfg.black_r =cfg.black_g =cfg.black_b = 0;
-cfg.dark_r  =cfg.dark_g  =cfg.dark_b  =20;
-cfg.medium_r=cfg.medium_g=cfg.medium_b=30;
-cfg.light_r =cfg.light_g =cfg.light_b =40;
-cfg.marked_r=cfg.marked_g=cfg.marked_b= 0; cfg.marked_r+=30;
+cfg.black.r =cfg.black.g =cfg.black.b = 0;
+cfg.dark.r  =cfg.dark.g  =cfg.dark.b  =20;
+cfg.medium.r=cfg.medium.g=cfg.medium.b=30;
+cfg.light.r =cfg.light.g =cfg.light.b =40;
+cfg.marked.r=cfg.marked.g=cfg.marked.b= 0; cfg.marked.r+=30;
 cfg.loop=0;
 cfg.mtype=MTYPE;
 for(f=0;f<256;f++) cfg.mode_allowed[f]=0;
 cfg.showxvpicdir=0;
-for(f=0;f<31;f++)
+for(f=0;f<=GLASTMODE;f++)
   cfg.mode_allowed[f]=vga_hasmode(f)?1:0;
 cfg.cmd[0]=0;
 cfg.errignore=0;
@@ -444,7 +489,7 @@ else
 
 /* all booleans are actually toggles. parsing happens after rc file reading.
  */
-int parsecommandline(int argc,char **argv)
+int parsecommandline(int argc,char *argv[])
 {
 int done;
 char buf[1024];
@@ -559,24 +604,24 @@ return(argc-optind);
 
 void usage_help()
 {
-printf("Zgv v%s - (c) 1993-1997 Russell Marks for improbabledesigns.",ZGV_VER);
+printf("Zgv v%s - (c) 1993-1998 Russell Marks for improbabledesigns.",ZGV_VER);
 
 printf(
 "
-usage: zgv [-bcghjJlktTz] [-a \"command\"] [-M mousetype] [-m modespec]
-	[-r seconds] [dirname | filename1 [... filenameN]]
+usage: zgv [-bcghijklpstTz] [-a \"command\"] [-J jpg_decmp_type] [-M mousetype]
+	[-m modespec] [-r seconds] [dirname | filename1 [... filenameN]]
 
 	-a	alternative command (see man page for details)
 	-b	blockcursor toggle, gives outline cursor not tacky 3d effect
 	-c	centre toggle, whether to centre pictures on the screen
-	        (defaults to on, so using '-c' will turn it off)
+	        (defaults to on, so using `-c' will turn it off)
 	-g	betterpgm toggle, whether to use 8-bit or high-colour modes
-	        for PGM files (defaults to on, so using '-g' will turn it off)
+	        for PGM files (defaults to on, so using `-g' will turn it off)
 	-h	give this usage help
         -i	ignore errors with GIF and PNG files and display whatever
         	part of the picture could be read anyway, if possible
-	-j	jpeg24bit toggle (see documentation, esp. 24bit.txt - defaults
-	        to on, so using '-j' will turn it off)
+	-j	jpeg24bit toggle (see documentation - defaults to on, so using
+        	`-j' will turn it off)
         -J	sets JPEG speed/quality tradeoff:
         		1 = floating-point - slow but accurate
                         2 = slow integer - faster but not as accurate
@@ -589,11 +634,11 @@ usage: zgv [-bcghjJlktTz] [-a \"command\"] [-M mousetype] [-m modespec]
 		Microsoft = `ms', Mouse Systems Corp = `msc',
 		MM Series = `mm', Logitech = `logi', BusMouse = `bm',
 		MSC 3-bytes = `sun', PS/2 = `ps2'
-	-m	startup mode; the 'modespec' should be in quotes but
-	        otherwise in the same format as the .zgvrc entry 'startmode'
+	-m	startup mode; the `modespec' should be in quotes but
+	        otherwise in the same format as the .zgvrc entry `startmode'
 	        e.g. \"640 480 8\"
         -p      turn off (or on) progress report when loading a single file
-        -r	re-read and redisplay picture every 'seconds' seconds
+        -r	re-read and redisplay picture every `seconds' seconds
         -s	print dimensions of section of picture being viewed to stdout
         	on exit
 	-t	thicktext toggle, makes the text a little bit clearer, kind of
