@@ -367,7 +367,7 @@ h=read_next_number(in);
 
 *width=w; *height=h;
 
-if(w==0 || h==0 || w>80 | h>60)
+if(w==0 || h==0 || w>80 || h>60)
   return(_PICERR_CORRUPT);
 
 /* for some reason, they have a maxval...!?
