@@ -1,12 +1,12 @@
 /* Zgv v2.7 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
  * Copyright (C) 1993-1995 Russell Marks. See README for license details.
  *
- * resizepic.c - the resizepic() routine, used by both zgv and tnpic.
+ * resizepic.c - the resizepic() routine.
  */
 
 
 #include <stdio.h>
-#include <malloc.h>
+#include <stdlib.h>
 #include "resizepic.h"
 
 
@@ -21,7 +21,7 @@ unsigned char *resizepic(unsigned char *theimage,
   unsigned char *palr,unsigned char *palg,unsigned char *palb,
   int width,int height,int *sw_ask,int *sh_ask)
 {
-register int a,b,x,yp,yw;
+int a,b,x,yp,yw;
 long y,sw,sh,lastyp;
 int c,pixwide,pixhigh;
 int scrnwide,scrnhigh;
@@ -62,7 +62,7 @@ if(bigimage)
     if(yp!=lastyp)
       {
       yw=y*width;
-      /* we try to resample a bit. get that pentium RSN :) */
+      /* we try to resample a bit. get that pentium RSN :-) */
       for(x=0;x<width;x++,yw++)
         {
         tr=tg=tb=tn=0;
