@@ -1,5 +1,5 @@
-/* Zgv v2.8 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
- * Copyright (C) 1993-1996 Russell Marks. See README for license details.
+/* Zgv v3.0 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993-1998 Russell Marks. See README for license details.
  *
  * vgadisp.c - vga specific display routines.
  */
@@ -8,7 +8,6 @@
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
-#include <malloc.h>
 #include <stdlib.h>
 #include <signal.h>
 #include <sys/file.h>
@@ -25,6 +24,7 @@
 #include "readpcx.h"
 #include "readngm.h"
 #include "readxvpic.h"
+#include "readmrf.h"
 #include "readnbkey.h"
 #include "helppage.h"
 #include "3deffects.h"
@@ -39,33 +39,34 @@ int curvgamode;
 int zoom,virtual;
 int vkludge;
 int pixelsize;		/* in bytes - 1 for 8-bit, 3 for 24-bit */
-static int isjpeg;
 int pic_incr;		/* after view, 0=stay, -1=prev, 1=next */
 
 char viewerhelp[][80]={
-  "? (question mark) \t this help page",
-  "5 to 8 \t\t mode change",
-  "F1 - F12  SF3 - SF10 \t SVGA mode change (try these first)",
-  "0 \t\t\t toggle virtual mode",
-  "v \t\t\t toggle smoothing in virtual/zoom modes",
-  "comma / dot \t contrast down/up (8-bit modes)",
-  "left / right chevron \t brightness down/up (8-bit modes)",
-  "asterisk \t\t reset contrast and brightness",
-  "z \t\t\t toggle zoomed mode",
-  "r \t\t\t rotate clockwise 90 degrees",
-  "m \t\t\t mirror",
-  "f \t\t\t flip",
-  "c \t\t\t toggle centering picture on screen",
-  "Esc or x \t\t exit the viewer",
+  "? (question mark)\\this help page",
+  "4 to 8\\mode change",
+  "F1-F12 SF3-SF10\\SVGA mode change (try these first)",
+  "v\\toggle smoothing in virtual/zoom modes",
+  "comma/dot\\contrast down/up (8-bit modes)",
+  "left/right chevron\\brightness down/up (8-bit modes)",
+  "asterisk\\reset contrast and brightness",
+  "z\\toggle zoomed mode",
+  "r\\rotate clockwise 90 degrees",
+  "m\\mirror",
+  "f\\flip",
+  "d\\double size of picture",
+  "D\\halve size of picture",
+  "Esc or x\\exit the viewer",
   " ",
   "hjkl or qaop keys scroll around the picture",
   "HJKL or QAOP (also the cursor keys) scroll in bigger steps",
   ""	/* end */
   };
 
-int mode_lookup[30]={
-  /* 0-4 unused */
-  0,0,0,0,0,
+int mode_lookup[50]={
+  /* 0-3 unused */
+  0,0,0,0,
+  /* 4 - 640x480x16, acts like a pseudo-8-bit mode */
+  G640x480x16,
   /* 5-8 - VGA modes */
   G320x200x256,G320x240x256,G320x400x256,G360x480x256,
   /* 9 - unused */
@@ -75,12 +76,18 @@ int mode_lookup[30]={
   G640x480x256, G800x600x256, G1024x768x256, G1280x1024x256,
   G320x200x32K, G320x200x64K, G320x200x16M,  G640x480x32K,
   G640x480x64K, G640x480x16M,
-  /* SF1-SF9 */
+  /* SF1-SF10 */
   G800x600x32K, G800x600x64K, G800x600x16M,  G1024x768x32K,
   G1024x768x64K,G1024x768x16M,G1280x1024x32K,G1280x1024x64K,
   G1280x1024x16M,
-  /* SF10 - unused */
-  0
+  0,		/* SF10 unused */
+  /* 30-49 - tab-prefixed SVGA modes */
+  /* 30-37 - F1-F8 */
+  G1152x864x256, G1152x864x32K, G1152x864x64K, G1152x864x16M,
+  G1600x1200x256,G1600x1200x32K,G1600x1200x64K,G1600x1200x16M,
+  /* 38-49 - F9-F10 and SF1-SF10; unused for now */
+  0,0,
+  0,0,0,0,0, 0,0,0,0,0
   };
 
 
@@ -107,11 +114,16 @@ int scaling=1,interp=0,inextpix=1;
 int palr[256],palg[256],palb[256];
 byte palr64[256],palg64[256],palb64[256];
 int palrgb[768];
+unsigned char dither16_greylookup[256];
+unsigned char dither16_rgb[768];
 
 int loading_file_type;
 int repeat_sig,first_repeat;
 int saved_px,saved_py;		/* for persistance in cfg.repeat_timer mode */
 
+/* grey values for rgb - NTSC uses these (adding up to 1000) */
+static int grey_red=299,grey_green=587,grey_blue=114;
+
 extern int zgv_ttyfd;
 int sgres;
 
@@ -146,8 +158,10 @@ void graphicson(void);
 void graphicsoff(void);
 int showgif(byte *palette);
 void repeat_sighandler(int foo);
+void setpalvec(int start,int num,int *pal);
 void redrawgif(int px, int py, int npx, int npy);
 void eventuallydrawscansegment(byte *ptr, int x, int y, int len);
+void dither16scansegment(unsigned char *ptr,int x,int y,int len);
 int getvpix(int px, int py, int x, int y);
 void drawzoomedgif(void);
 void fx_mirror(void);
@@ -174,7 +188,6 @@ int result=0;
 PICINFO ginfo;
 byte *palette;
 
-
 first_repeat=1;
 
 do
@@ -241,11 +254,12 @@ do
         else
           if(result!=_PIC_OK) { free(theimage); theimage=NULL; }
         
+      if(result!=_PIC_OK) result=_PICERR_SHOWN_ALREADY;
       break;
 #endif /* PNG_SUPPORT */
       
     case _IS_PNM: case _IS_BMP: case _IS_TGA: case _IS_PCX:
-    case _IS_NGM: case _IS_XVPIC:
+    case _IS_NGM: case _IS_XVPIC: case _IS_MRF:
       if(show_dont_tell) pixelsize=has_highcolour_mode()?3:1;
       switch(loading_file_type)
         {
@@ -267,6 +281,9 @@ do
 	case _IS_XVPIC:
           result=read_xvpic(giffn,howfarfunc,&theimage,&palette,&pixelsize,
                   &ginfo); break;
+	case _IS_MRF:
+          result=read_mrf_file(giffn,howfarfunc,&theimage,&palette,&pixelsize,
+                  &ginfo); break;
         }
       
       width=ginfo.width;
@@ -351,13 +368,22 @@ vga_modeinfo *vminfo;
  * - we then try to match any 8-bit modes for EIGHT_BIT, or match
  *   24/16/15-bit modes (in that order) for HIGHCOLOUR.
  * - of course, the mode has to be either 8 or 15/16/24-bit as reqd.
+ *
+ * no support in this for 640x480x4 (16-colour) - because that's
+ * a really weird special-case mode, you have to explicitly select it
+ * if you want it. However, it sticks in 640x480x4 if selected and
+ * modetype is FIX_TO_EIGHT_BIT, because it'd be totally annoying
+ * otherwise.
  */
 
+if(curvgamode==G640x480x16 && modetype==FIX_TO_EIGHT_BIT) return;
+
 newmode=-1; pixdiff=(1<<30);
 numpixels=get_mode_width(curvgamode)*get_mode_height(curvgamode);
 
 /* account for effect of 'virtual' */
-if((curvgamode==7)||(curvgamode==8)) numpixels*=2;
+if((curvgamode==G320x400x256)||(curvgamode==G360x480x256))
+  numpixels*=2;
 
 /* we check 0 - 255 at the most */
 for(f=0;f<256;f++)
@@ -379,7 +405,8 @@ for(f=0;f<256;f++)
   if(mode_ok==0) continue;
   
   newnp=(vminfo->width)*(vminfo->height);
-  if((f==7)||(f==8)) newnp*=2;   /* account for effect of 'virtual' */
+  if((f==G320x400x256)||(f==G360x480x256))
+    newnp*=2;   /* account for effect of 'virtual' */
   newdiff=numpixels-newnp;
   if(newdiff<0) newdiff=(-newdiff);
   
@@ -394,7 +421,10 @@ for(f=0;f<256;f++)
  * out for us (hence the '<=' above).
  */
 curvgamode=newmode;
-if((curvgamode==7)||(curvgamode==8)) virtual=1; else virtual=0;
+if((curvgamode==G320x400x256)||(curvgamode==G360x480x256))
+  virtual=1;
+else
+  virtual=0;
 }
 
 
@@ -422,31 +452,25 @@ switch(loading_file_type)
     aborted_file_png_cleanup();
     break;
 #endif
-#ifdef NGM_SUPPORT
   case _IS_NGM:
     aborted_file_ngm_cleanup();
     break;
-#endif
   case _IS_XVPIC:
     fprintf(stderr,"aborted xvpic load: can't happen!\n");
     exit(1);
   case _IS_PCX:
     aborted_file_pcx_cleanup();
     break;
+  case _IS_MRF:
+    aborted_file_mrf_cleanup();
+    break;
   }
 }
 
 
-int is_this_file_jpeg()
-{
-return(isjpeg);
-}
-
-
 void makerealpal()
 {
 int f;
-int red=299, green=587, blue=114;	/* NTSC uses these */
 int r,rlo,g,glo,b,blo;
 int gval;
 
@@ -458,7 +482,7 @@ for(f=0;f<256;f++)
 
   if(cfg.gamma)
     {
-    gval=(rlo*red+glo*green+blo*blue)/4;
+    gval=(rlo*grey_red+glo*grey_green+blo*grey_blue)/4;
     /* so gval is sub-value, 0<=gval<1000. (really <750... :-( )
      * possible emulated sub-values are...
      * 0, 114 (b), 299 (r), 413 (r+b), 587 (g), 701 (g+b), 886 (r+g)
@@ -560,6 +584,8 @@ void graphicson()
 {
 int x;
 
+msgbox_draw_ok=0;
+
 vga_lockvc();
 if((vga_hasmode(curvgamode))&&(cfg.mode_allowed[curvgamode]))
   vga_setmode(curvgamode);
@@ -567,26 +593,50 @@ else
   {
   /* we haven't? Aaargh!!! Ok, use 640x480 or 360x480 instead. */
   if((vga_hasmode(G640x480x256))&&(cfg.mode_allowed[G640x480x256]))
-    {
     vga_setmode(curvgamode=G640x480x256);
-    virtual=0;
-    }
   else
     {
     if((vga_hasmode(G360x480x256))&&(cfg.mode_allowed[G360x480x256]))
-      {
       vga_setmode(curvgamode=G360x480x256);
-      virtual=1;
-      }
     else
-      {
       /* *must* have 320x200 (see rcfile.c for more info) */
       vga_setmode(curvgamode=G320x200x256);
-      virtual=0;
-      }
     }
   }
 
+virtual=0;
+if(curvgamode==G320x400x256 || curvgamode==G360x480x256)
+  virtual=1;
+
+if(curvgamode==G640x480x16)
+  {
+  int paltmp[16*3],f,r,g,b;
+  
+  if(cfg.viewer16col)
+    {
+    /* colour */
+    for(f=r=0;r<2;r++)
+      for(g=0;g<2;g++)
+        for(b=0;b<2;b++,f+=3)
+          paltmp[f]=r*63,paltmp[f+1]=g*63,paltmp[f+2]=b*63;
+    /* FWIW, zero colours 8..15. It may seem wasteful not using these,
+     * but there's no way I can see to use them as part of an ordered
+     * dither which doesn't look awful (and error-diffused dithering
+     * is too slow).
+     */
+    for(;f<16*3;f+=3)
+      paltmp[f]=paltmp[f+1]=paltmp[f+2]=0;
+    }
+  else
+    {
+    /* 16-grey palette */
+    for(f=0;f<16;f++)
+      paltmp[f*3]=paltmp[f*3+1]=paltmp[f*3+2]=(f*63)/15;
+    }
+  
+  vga_setpalvec(0,16,paltmp);
+  }
+
 if(pixelsize!=1) gl_setcontextvga(curvgamode);
 
 scrnwide=get_mode_width(curvgamode);
@@ -595,9 +645,9 @@ scrnpixelsize=get_mode_pixelbytes(curvgamode);
 scrncols=get_mode_numcols(curvgamode);
 
 if(virtual)
-  scrnwide*=2;
+  scrnwide*=2,scaling=1;
 if(pixelsize==1)
-  vga_setpalvec(0,256,palrgb);
+  setpalvec(0,256,palrgb);
 if(vkludge)
   {
   for(x=0;x<65536;x++) vkcache[x].rgb=-1;
@@ -626,6 +676,7 @@ int f;
 extern int x,y,mx,my, has_mouse, mfd;
 int left_button=0, right_button=0, butstatep=0;
 struct ms_event ev;
+int mode_prefix=0;
 
 wait_for_foreground();
 
@@ -636,6 +687,9 @@ if(numcols<256)
   {
   for(f=numcols;f<numcols+64;f++)
     palette[f]=palette[256+f]=palette[512+f]=((f-numcols)<<2);
+  /* get rid of any junk entries which vkludge might try and use */
+  for(;f<256;f++)
+    palette[f]=palette[256+f]=palette[512+f]=0;
   }
 
 pic_incr=repeat_sig=0;
@@ -680,7 +734,7 @@ redrawgif(px,py,npx,npy);
 
 if(has_mouse && mfd != -1)
   {
-  int c;
+  unsigned char c;
   while(read(mfd,&c,1)==1);	       /* clear the read buffer */
   }
 
@@ -713,47 +767,51 @@ while(!quitshow && repeat_sig!=1)
       
   /* svga modes - note that F11=shift-F1 and F12=shift-F2 */
   if(((key>=RK_F1)&&(key<=RK_F12))||((key>=RK_SHIFT_F3)&&(key<=RK_SHIFT_F10)))
-    /* the RK_F?'s are defined such that this sort of thing is ok to do: */
-    /* (here comes the 'if' statement from hell...) */
-    if((vga_hasmode(key-RK_F1+10))&&(cfg.mode_allowed[key-RK_F1+10])&&
-       (((pixelsize==1)&&(get_mode_numcols(key-RK_F1+10)==256))||
-        ((pixelsize==3)&&(get_mode_numcols(key-RK_F1+10)>256))))
-      {
-      curvgamode=mode_lookup[key-RK_F1+10];   /* so F1 = mode 10 */
-      virtual=0;
-      graphicson();
-      redraw=1;
-      }
-
-  if(key=='0' || (key>='5' && key<='8'))
     {
-    if(key=='0')
+    /* the RK_F?'s are defined such that this sort of thing is ok to do: */
+    int idx=key-RK_F1+10,modenum;   /* so F1 = mode 10 */
+    
+    if(mode_prefix) idx+=20;		/* past normal F1..F10 and SF1..SF10 */
+    modenum=mode_lookup[idx];
+    mode_prefix=0;
+    
+    /* check the mode is ok */    
+    if(vga_hasmode(modenum) && cfg.mode_allowed[modenum] &&
+       ((pixelsize==1 && get_mode_numcols(modenum)<=256) ||
+        (pixelsize==3 && get_mode_numcols(modenum)> 256)))
       {
-      virtual=(virtual==1)?0:1;
+      curvgamode=modenum;
       graphicson();
       redraw=1;
       }
-    else
+    }
+
+  if(key>='4' && key<='8')
+    {
+    if((vga_hasmode(key-48))&&(cfg.mode_allowed[key-48])&&
+       (((pixelsize==1)&&(get_mode_numcols(key-48)<=256))||
+        ((pixelsize==3)&&(get_mode_numcols(key-48)>256))))
       {
-      if((vga_hasmode(key-48))&&(cfg.mode_allowed[key-48])&&
-         (((pixelsize==1)&&(get_mode_numcols(key-48)==256))||
-          ((pixelsize==3)&&(get_mode_numcols(key-48)>256))))
-        {
-        curvgamode=key-48;
-        if((curvgamode==7)||(curvgamode==8)) virtual=1; else virtual=0;
-        if(virtual) scaling=1;
-        graphicson();
-        redraw=1;
-        }
+      curvgamode=key-48;
+      graphicson();
+      redraw=1;
       }
+    
+    mode_prefix=0;
     }
   else
+    {
+    if(key) mode_prefix=0;
+    
     switch(key)
       {
+      case 9:		/* tab */
+        mode_prefix=1;
+        break;
       case 'g':
         cfg.gamma=!cfg.gamma;
         if(pixelsize==1)
-          filterpal(palette),vga_setpalvec(0,256,palrgb);
+          filterpal(palette),setpalvec(0,256,palrgb);
         break;
       case 's': case 'd':
         if(virtual) break;
@@ -806,15 +864,15 @@ while(!quitshow && repeat_sig!=1)
         px+=scrnwide*9/10; break;
       case 'v': vkludge=((vkludge==1)?0:1); redraw=1; graphicson(); break;
       case ',': contrast-=0.05;
-        filterpal(palette); vga_setpalvec(0,256,palrgb); break;
+        filterpal(palette); setpalvec(0,256,palrgb); break;
       case '.': contrast+=0.05;
-        filterpal(palette); vga_setpalvec(0,256,palrgb); break;
+        filterpal(palette); setpalvec(0,256,palrgb); break;
       case '<': brightness-=10;
-        filterpal(palette); vga_setpalvec(0,256,palrgb); break;
+        filterpal(palette); setpalvec(0,256,palrgb); break;
       case '>': brightness+=10;
-        filterpal(palette); vga_setpalvec(0,256,palrgb); break;
+        filterpal(palette); setpalvec(0,256,palrgb); break;
       case '*': contrast=1.0; brightness=0;
-        filterpal(palette); vga_setpalvec(0,256,palrgb); break;
+        filterpal(palette); setpalvec(0,256,palrgb); break;
       case 'q': case 'k': py-=MOVSTP; break;
       case 'a': case 'j': py+=MOVSTP; break;
       case 'o': case 'h': px-=MOVSTP; break;
@@ -851,11 +909,20 @@ while(!quitshow && repeat_sig!=1)
 	pic_incr=1; quitshow=2; break;
       case ' ':
 	pic_incr=1; quitshow=3; break;
-       case 127: case 8:
+      case 127: case 8:
 	pic_incr=-1; quitshow=2; break;
+      case 'C':
+        if(curvgamode==G640x480x16)
+          {
+          cfg.viewer16col=!cfg.viewer16col;
+          graphicson();		/* palette has to change */
+          redraw=1;
+          }
+        break;
       case RK_ESC: case 'x':
         quitshow=1;
       }
+    }
       
   if(!zoom)
     {
@@ -906,6 +973,45 @@ repeat_sig=1;
 }
 
 
+void setpalvec(int start,int num,int *pal)
+{
+/* only relevant for 8-bit modes */
+if(pixelsize==1)
+  if(curvgamode!=G640x480x16)
+    vga_setpalvec(start,num,pal);
+  else
+    {
+    int f,c;
+    unsigned char *dptr=dither16_rgb+start*3;
+    int *sptr=pal+start*3;
+    
+    for(f=start;f<start+num;f++)
+      {
+      if(cfg.viewer16col)
+        {
+        /* add contrast to compensate for dithering (kind of) */
+        c=(*sptr++-32)*4/3+32;
+        if(c<0) c=0; if(c>63) c=63;
+        *dptr++=c;
+        c=(*sptr++-32)*4/3+32;
+        if(c<0) c=0; if(c>63) c=63;
+        *dptr++=c;
+        c=(*sptr++-32)*4/3+32;
+        if(c<0) c=0; if(c>63) c=63;
+        *dptr++=c;
+        }
+      else
+        {
+        c=(pal[f*3]*grey_red+pal[f*3+1]*grey_green+
+           pal[f*3+2]*grey_blue)/1000;
+        if(c>63) c=63;
+        dither16_greylookup[f]=c;
+        }
+      }
+    }
+}
+
+
 /* this routine is getting ridiculous */
 void redrawgif(int px,int py,int npx,int npy)
 {
@@ -935,7 +1041,7 @@ else
   if(swidth-px<scrnwide) xdim=swidth-px; else xdim=scrnwide;
   if((py>=sheight)||(px>=swidth)) return;
   /* hopefully the following is fairly quick as fewer ifs... ? */
-  if((virtual)&&(pixelsize==1))
+  if(virtual)		/* 320x400 or 360x480 */
     {
     if((realline=calloc(1,scrnwide))==NULL) return;
     for(y=0;(y<sheight-py)&&(y<scrnhigh);y++)
@@ -1056,7 +1162,7 @@ else
       }
     free(realline);
     }
-  else
+  else		/* if not scaled and not zoomed... */
     {
     switch(scrnpixelsize)
       {
@@ -1069,20 +1175,20 @@ else
         /* for 320x200 (I think), 320x240, 320x400, 360x480 we need the
          * length to be a multiple of 8. :-( We allocate an image-width
          * line (blanked), and display via that.
+         * (320x400/360x480 were dealt with earlier though; can only
+         * be 320x200/320x240 if it gets here.)
          */
-        if(curvgamode==G320x200x256 || curvgamode==G320x240x256 ||
-           curvgamode==G320x400x256 || curvgamode==G360x480x256)
+        if(curvgamode==G320x200x256 || curvgamode==G320x240x256)
           {
-          unsigned char *hack;
+          static unsigned char hack[320];
           
-          if((hack=malloc(scrnwide))==NULL) return;
+          memset(hack,0,320);
           for(y=0;(y<height-py)&&(y<scrnhigh);y++)
             {
             /* assume pixelsize==1, because it must be */
-            memcpy(hack,theimage+((py+y)*width+px),xdim);
-            vga_drawscansegment(hack,x_add,y+y_add,xdim);
+            memcpy(hack+x_add,theimage+((py+y)*width+px),xdim);
+            vga_drawscansegment(hack,0,y+y_add,scrnwide);
             }
-          free(hack);	/* bit like zgv, really :-) */
           }
         else
           for(y=0;(y<height-py)&&(y<scrnhigh);y++)
@@ -1092,8 +1198,6 @@ else
       }
     }
   }
-
-return;
 }
 
 
@@ -1102,7 +1206,10 @@ void eventuallydrawscansegment(byte *ptr,int x,int y,int len)
 static unsigned short buf[32768];
 
 if(scrnpixelsize!=2)
-  vga_drawscansegment(ptr,x,y,len);
+  if(curvgamode==G640x480x16)
+    dither16scansegment(ptr,x,y,len);
+  else
+    vga_drawscansegment(ptr,x,y,len);
 else
   {
   unsigned short *bufref=buf;
@@ -1120,6 +1227,65 @@ else
 }
 
 
+/* only called for 640x480x4-bit */
+void dither16scansegment(unsigned char *ptr,int x,int y,int len)
+{
+static unsigned char scanline[640];
+unsigned char *lptr=scanline;
+int f;
+
+if(!cfg.viewer16col)
+  {
+  /* greyscale, but dithered between the scales;
+   * dithering slows it down more than a direct mapping would,
+   * but it just looks too nasty without it.
+   */
+  static unsigned char ditherdata[4]=
+    {
+    1,4,
+    3,2
+    };
+  unsigned char c,*dithptr=ditherdata+(y&1)*2+(x&1);
+  int xm=2-(x&1);
+  
+  for(f=0;f<len;f++)
+    {
+    if((c=dither16_greylookup[ptr[f]])>=60)
+      c=60;	/* saves needing a range check after :-) */
+    
+    *lptr++=((c>>2)+((c&3)>=*dithptr++));
+    if(!(--xm))
+      dithptr-=(xm=2);
+    }
+  }
+else
+  {
+  /* colour dither */
+  static unsigned char ditherdata[16]=
+    {
+    1*4,13*4, 4*4,15*4,
+    9*4, 5*4,12*4, 8*4,
+    3*4,15*4, 2*4,14*4,
+    7*4,10*4, 6*4,11*4
+    };
+  unsigned char *dithptr=ditherdata+(y&3)*4+(x&3);
+  unsigned char *palptr,d;
+  int xm=4-(x&3);
+  
+  for(f=0;f<len;f++)
+    {
+    d=*dithptr++;
+    palptr=dither16_rgb+ptr[f]*3;
+    *lptr++=(((*palptr>=d)*4)|((palptr[1]>=d)*2)|(palptr[2]>=d));
+    if(!(--xm))
+      dithptr-=(xm=4);
+    }
+  }
+
+vga_drawscansegment(scanline,x,y,len);
+}
+
+
 int getvpix(int px,int py,int x,int y)
 {
 int p1,p2,r,g,b;
@@ -1142,18 +1308,16 @@ else
 /* this routine is nasty writ big, but about as quick as I can manage */
 void drawzoomedgif()
 {
-register int a,b,x,yp,yw;
+int a,b,x,yp,yw;
 long y,sw,sh,lastyp;
 int xoff,yoff;
 int x_add,y_add;	/* for centering */
-int c=0,cr=0,cg=0,cb=0,pixwide,pixhigh;
+int pixwide,pixhigh;
 int bigimage;
-byte *rline;
+byte *rline,*sptr,*dptr;
 int tmp1,tmp2,tr,tg,tb,tn;
 
-if((rline=malloc(scrnwide*pixelsize))==NULL) return;
-for(x=0;x<scrnwide*pixelsize;x++)
-  rline[x]=0;
+if((rline=calloc(scrnwide*pixelsize,1))==NULL) return;
 
 /* try landscapey */
 sw=scrnwide; sh=(scrnwide*height)/width;
@@ -1168,10 +1332,10 @@ x_add=y_add=0;
 if(cfg.centreflag)
   {
   if(sw<scrnwide)
-    x_add=(scrnwide-sw)>>1;
+    x_add=((scrnwide-sw)>>1);
   
   if(sh<scrnhigh)
-    y_add=(scrnhigh-sh)>>1;
+    y_add=((scrnhigh-sh)>>1);
   
   if(virtual) x_add>>=1;
   }
@@ -1193,8 +1357,9 @@ if(bigimage)
     if(yp!=lastyp)
       {
       yw=y*width;
-      if(vkludge) /* we try to resample a bit. get that pentium RSN :-) */
+      if(vkludge)
         {
+        /* we try to resample it a bit */
         for(x=0;x<width;x++,yw++)
           {
           tr=tg=tb=tn=0;
@@ -1228,7 +1393,7 @@ if(bigimage)
             }
           }
         }
-      else
+      else	/* not vkludge... */
         for(x=0;x<width;x++,yw++)
           {
           if(scrncols<=256)
@@ -1236,59 +1401,48 @@ if(bigimage)
           else
             memcpy(rline+(x*sw)/width*3,theimage+yw*3,3);
           }
-      eventuallydrawscansegment(rline,x_add,yp+y_add,scrnwide*pixelsize);
+      eventuallydrawscansegment(rline,x_add,yp+y_add,sw*pixelsize);
       lastyp=yp;
       }
     }
   free(rline);
   }
 else
-  /* well, we need to fill in the gaps because it's been made bigger. */
   {
-  /* a pixel from the original is now pixwide x pixhigh */
-  pixwide=(int) ( (((float)sw)/((float)width )) +1.0 );
-  pixhigh=(int) ( (((float)sh)/((float)height)) +1.0 );
-
-  free(rline);  /* get rid of that... */
-    /* and get a screen's worth... hahahahahahahahaha er, sorry. */
-  if((rline=malloc(pixelsize*scrnwide*(scrnhigh+pixhigh)))==NULL) return;
-    /* notice that we allow for a screwy 'pixhigh' value to avoid
-       any nasty old segfaults! oh, now we wipe it: */
-  for(x=0;x<scrnwide*scrnhigh*pixelsize;x++) rline[x]=0;
-  
+  /* well, we need to fill in the gaps because it's been made bigger.
+   * However, it makes more sense to look at it the other way around.
+   * So we don't just make small pixels bigger; instead, we go through
+   * each of the pixels in the scaled-up image and copy an appropriate
+   * pixel from the original (smaller) image. This is a lot simpler
+   * and cleaner than the more obvious approach - and I speak from
+   * experience. :-)
+   */
   if(virtual)
-    sw>>=1;   /* leave pixwide though as it *might* miss with >> */
-  for(y=0;y<height;y++)
+    sw>>=1;
+  for(y=0;y<sh;y++)
     {
-    yoff=(y*sh)/height;
-    yw=y*width*pixelsize;
-    for(x=0;x<width;x++,yw+=pixelsize)
+    yoff=(y*height)/sh; 
+    if(yoff>=height) yoff=height-1;
+    dptr=rline;
+   
+    for(x=0;x<sw;x++)
       {
+      xoff=(x*width)/sw;
+      if(xoff>=width) xoff=width-1;
+      
       if(scrncols<=256)
-        c=*(theimage+yw);
+        /* image must be 8-bit */
+        *dptr++=*(theimage+yoff*width+xoff);
       else
         {
-        cb=*(theimage+yw); cg=*(theimage+yw+1); cr=*(theimage+yw+2);
+        sptr=theimage+3*(yoff*width+xoff);
+        *dptr++=*sptr; *dptr++=sptr[1]; *dptr++=sptr[2];
         }
-      xoff=(x*sw)/width;
-      for(b=0;b<pixhigh && yoff+b<scrnhigh;b++)
-        for(a=0;a<pixwide && xoff+a<scrnwide;a++)
-          if(scrncols<=256)
-            *(rline+(yoff+b)*scrnwide+xoff+a)=c;
-          else
-            {
-            tmp1=3*((yoff+b)*scrnwide+xoff+a);
-            *(rline+tmp1)=cb;
-            *(rline+tmp1+1)=cg;
-            *(rline+tmp1+2)=cr;
-            }
       }
-    eventuallydrawscansegment(rline+pixelsize*y*scrnwide,
-    	x_add,y+y_add,scrnwide*pixelsize); /* really pessimistic */
+    
+    /* draw the line */
+    eventuallydrawscansegment(rline,x_add,y+y_add,sw*pixelsize);
     }
-  for(;y<scrnhigh;y++)          /* and do the rest */
-    eventuallydrawscansegment(rline+pixelsize*y*scrnwide,
-    				x_add,y+y_add,scrnwide*pixelsize);
   free(rline);
   }
 }
@@ -1413,8 +1567,8 @@ int closest(int r,int g,int b)
 {
 int idx,rgb;
 byte *pr,*pg,*pb;
-register byte distnum;
-register int xr,xg,xb,dist,distquan,f,checknumcols;
+byte distnum;
+int xr,xg,xb,dist,distquan,f,checknumcols;
 
 rgb=((b<<12)|(g<<6)|r);
 idx=rgb&0xffff;
