@@ -1,5 +1,5 @@
-/* Zgv v2.8 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
- * Copyright (C) 1993-1997 Russell Marks.
+/* Zgv v3.0 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993-1998 Russell Marks.
  *
  * zgv.c - This provides the zgv file selector, and interfaces to the
  *         vga display routines (vgadisp.c)
@@ -52,25 +52,23 @@
 #include "zgvlogopck.h"
 
 
-char zgv_pngerr[128];	/* for error messages from pnglib */
-
 char zgvhelp[][80]={
-  " ", " ",
-  "? (question mark) \t this help page",
   " ",
-  "up (also k or q) \t move file selection cursor up",
-  "down (also j or a) \t move file selection cursor down",
-  "left (also h or o) \t move file selection cursor left",
-  "right (also l or p) \t move file selection cursor right",
+  "? (question mark)\\this help page",
   " ",
-  "Enter \t\t display file or change directory",
+  "up (also k or q)\\move file selection cursor up",
+  "down (also j or a)\\move file selection cursor down",
+  "left (also h or o)\\move file selection cursor left",
+  "right (also l or p)\\move file selection cursor right",
   " ",
-  "v \t\t\t visual selection mode on/off",
-  "u \t\t\t create/update thumbnails in visual mode",
-  "Ctrl-R \t\t update directory listing / redraw screen",
-  "D or Delete \t\t delete file",
+  "Enter\\display file or change directory",
   " ",
-  "Esc or x \t\t exit zgv",
+  "v\\visual selection mode on/off",
+  "u\\create/update thumbnails in visual mode",
+  "Ctrl-R\\update directory listing / redraw screen",
+  "D or Delete\\delete file",
+  " ",
+  "Esc or x\\exit zgv",
   ""
   };
 
@@ -87,13 +85,14 @@ char zgvhelp[][80]={
 #define MIDGREY		0
 #define MARKEDCOL	14
 
-#define GDFSIZ			3
+#define GDFYBIT			18
 #define BARWIDTH		138
 #define BAR_RESTRICT_WIDTH	BARWIDTH-6
 
 #define COLS		600
 #define MDIRSIZ		4096
 #define YSIZ		((395-yofs)/barheight)
+#define MPPOSSIZ	256
 
 /* number of greyscales *less one* used in selector when in 640x480x16 mode */
 #define GREY_MAXVAL	10
@@ -118,12 +117,23 @@ int idx_light,idx_medium,idx_dark,idx_black,idx_marked;
 int updating_index=0;
 int v256=1;	/* 1=256-col selector, 0=16-col selector */
 
+int gdfsiz=3;		/* size of filename text in selector */
+int gdfofs=0;
+
 int barheight,yofs;	/* height of cursor, and offset of selection bit */
 
 /* XXX this sgres stuff is bogus in the extreme */
 extern int sgres;
 int has_mouse=0;
 
+int current_colour=0;	/* see zgv.h */
+
+struct pastpos_t
+  {
+  int dev,inode,curent,startfrom;
+  } pastpos[MPPOSSIZ];
+
+
 
 /* prototypes */
 int main(int argc, char **argv);
@@ -131,6 +141,8 @@ void openstdin_nonblocking(void);
 void load_one_file(char *filename);
 void copyfromconfig(void);
 void mainloop(void);
+void new_pastpos(int curent, int startfrom);
+void get_pastpos(int *curentp,int *startfromp);
 int load_single_file(int curent, int do_howfar);
 void load_tagged_files(void);
 int permissiondenied(char *fname);
@@ -141,6 +153,7 @@ void smallhowfar(int sofar, int total);
 void showbar(int entnum, int selected, int startfrom);
 int centreseltxt(int x, int fsiz, char *str);
 void showgifdir(int startfrom, int unshow, int drawdirmsg);
+int ispicture(char *filename);
 void readgifdir(void);
 int gcompare(void *gn1, void *gn2);
 void prettyfile(char *buf, struct GDIR *gifdptr);
@@ -153,7 +166,8 @@ int delete_file(char *filename);
 void xv332_how_far(int sofar, int total);
 void clear_xvpic(int xpos, int ypos);
 int makexv332(char *filename, char *xvpicfn, unsigned int howfar);
-void update_xvpics(void);
+int makedirxv332(char *filename, char *xvpicfn, unsigned int howfar);
+void update_xvpics(int do_dirs_instead);
 int fixvt(void);
 void switchback(void);
 void greyfix332(unsigned char **image, int w, int h, int xw7, int w8);
@@ -162,7 +176,7 @@ void ctrlc(int foo);
 
 
 
-int main(int argc,char **argv)
+int main(int argc,char *argv[])
 {
 int argsleft;
 
@@ -182,9 +196,9 @@ vga_init();
 /* root permissions should now have been ditched */
 
 pixelsize=1;
-getconfig();				/* see rcfile.c... */
+getconfig();				/* see rcfile.c...   */
 argsleft=parsecommandline(argc,argv);	/* ...for these two. */
-    
+
 copyfromconfig();
     
 if(cfg.mtype != NO_MOUSE)
@@ -321,10 +335,14 @@ void mainloop()
 {
 int quit,key,curent; /* quit, key pressed, current entry */
 int oldent,startfrom,oldstart,f,markchange;
-extern int x,y,mx,my;
+extern int x,y,mx,my;	/* ... don't blame me, I didn't write that ;-) -rjm */
 int left_button=0, right_button=0, butstatep=0,px,py;
 struct ms_event ev;
 
+/* black out past-positions array */
+for(f=0;f<MPPOSSIZ;f++)
+  pastpos[f].dev=-1,pastpos[f].inode=-1;
+
 quit=0; curent=1; oldent=1;
 startfrom=1;
 
@@ -406,12 +424,21 @@ while(!quit)
       if(strstr(cdir,"/.xvpics")==NULL && cfg.xvpic_index)
         {
         updating_index=1;
-        update_xvpics();
+        update_xvpics(0);
         updating_index=0;
         redrawall(curent,startfrom);
         }
       }
       break;
+    case 'd':
+      if(cfg.xvpic_index)
+        {
+        updating_index=1;
+        update_xvpics(1);
+        updating_index=0;
+        redrawall(curent,startfrom);
+        }
+      break;
     case 'v':
       cfg.xvpic_index=!cfg.xvpic_index;
       redrawall(curent,startfrom);
@@ -480,12 +507,12 @@ while(!quit)
       if(curent<gifdirsiz) curent++;
       break;
     case 'L':	/* toggle loop or not in slideshow */
-      cfg.loop=1 - cfg.loop;
+      cfg.loop=!cfg.loop;
       if(cfg.loop) 
-         msgbox(zgv_ttyfd,"Loop in slideshow.",
+         msgbox(zgv_ttyfd,"Loop in slideshow",
                 MSGBOXTYPE_OK,idx_light,idx_dark,idx_black);
       else
-         msgbox(zgv_ttyfd,"No looping in slideshow.",
+         msgbox(zgv_ttyfd,"No looping in slideshow",
                 MSGBOXTYPE_OK,idx_light,idx_dark,idx_black);
       redrawall(curent,startfrom);
       break;
@@ -525,11 +552,18 @@ while(!quit)
           
         if(gifdir[curent].isdir)
           {
+          new_pastpos(curent,startfrom);
           showbar(curent,0,startfrom);
           showgifdir(startfrom,1,1);
           chdir(gifdir[curent].name);
           readgifdir();
-          showgifdir(1,0,1); curent=oldent=startfrom=1;
+          get_pastpos(&curent,&startfrom);
+          /* if these are off the end, it's clearly out of date (dir
+           * must have changed) so might as well just start at top.
+           */
+          if(curent>gifdirsiz || startfrom>gifdirsiz) curent=startfrom=1;
+          oldent=curent;
+          showgifdir(startfrom,0,1);
           showbar(curent,1,startfrom);
           break;
           }
@@ -565,6 +599,14 @@ while(!quit)
       load_tagged_files();
       redrawall(curent,startfrom);
       break;
+    
+    case ':':		/* give full filename, for you nethack fans :-) */
+    case ';':		/* (also allow semicolon, might as well) */
+      msgbox(zgv_ttyfd,gifdir[curent].name,MSGBOXTYPE_OK,
+         	idx_light,idx_dark,idx_black);
+      redrawall(curent,startfrom);
+      break;
+    
     case 'x': case RK_ESC:
       quit=1;
     }
@@ -592,6 +634,52 @@ while(!quit)
 }
 
 
+/* add new pastpos[0], shifting down all the rest of the entries. */
+void new_pastpos(int curent,int startfrom)
+{
+struct stat sbuf;
+int f;
+
+if(cfg.forgetoldpos) return;
+
+for(f=MPPOSSIZ-1;f>0;f--)
+  {
+  pastpos[f].dev      =pastpos[f-1].dev;
+  pastpos[f].inode    =pastpos[f-1].inode;
+  pastpos[f].curent   =pastpos[f-1].curent;
+  pastpos[f].startfrom=pastpos[f-1].startfrom;
+  }
+
+if(stat(".",&sbuf)==-1) return;
+
+pastpos[0].dev      =sbuf.st_dev;
+pastpos[0].inode    =sbuf.st_ino;
+pastpos[0].curent   =curent;
+pastpos[0].startfrom=startfrom;
+}
+
+
+/* return curent from pastpos[] entry matching current directory,
+ * or if none match return 1.
+ */
+void get_pastpos(int *curentp,int *startfromp)
+{
+struct stat sbuf;
+int f;
+
+*curentp=*startfromp=1;
+
+if(cfg.forgetoldpos || stat(".",&sbuf)==-1) return;
+
+for(f=0;f<MPPOSSIZ;f++)
+  if(pastpos[f].inode==sbuf.st_ino && pastpos[f].dev==sbuf.st_dev)
+    {
+    *curentp=pastpos[f].curent;
+    *startfromp=pastpos[f].startfrom;
+    return;
+    }
+}
+
 
 int load_single_file(int curent,int do_howfar)
 {
@@ -682,8 +770,6 @@ do {
   dohf=0;
   }
  } while (cfg.loop && sgres != 1);
-
-return;
 }
 
 
@@ -718,7 +804,7 @@ draw3dbox(100,220,540,260,2,1, idx_light,idx_dark);
 draw3dbox(109,229,531,251,1,0, idx_light,idx_dark);
 
 howfar_upto=0;
-drawtext3d(250,235,2,"Decompressing - please wait",0,
+drawtext3d(255-10*cfg.linetext,235,2,"Decompressing - please wait",0,
 					idx_light,idx_dark,idx_black);
 }
 
@@ -780,7 +866,7 @@ ypos=fwinypos(entnum-startfrom+1);
 prettyfile(ctmp,&(gifdir[entnum]));
 
 set_max_text_width(BAR_RESTRICT_WIDTH);
-xt=cfg.xvpic_index?centreseltxt(xpos,GDFSIZ,ctmp):xpos+3;
+xt=cfg.xvpic_index?centreseltxt(xpos,gdfsiz,ctmp):xpos+3;
 if(cfg.blockcursor)
   {
   /* block-style cursor - hopefully easier to read/see. */
@@ -792,7 +878,7 @@ if(cfg.blockcursor)
   
   /* in case the last file was just marked/unmarked */
   vga_setcolor(gifdir[entnum].marked?idx_marked:idx_black);
-  vgadrawtext(xt,ypos+3,GDFSIZ,ctmp);
+  vgadrawtext(xt,ypos+3+gdfofs,gdfsiz,ctmp);
   }
 else
   {
@@ -800,33 +886,32 @@ else
     {
     draw3dbox(xpos,ypos,xpos+BARWIDTH-1,ypos+barheight-1,1,1,
     	idx_light,idx_dark);
-    drawtext3d(xt,ypos+3,GDFSIZ,ctmp,0, idx_light,idx_dark,
+    drawtext3d(xt,ypos+3+gdfofs,gdfsiz,ctmp,0, idx_light,idx_dark,
     		gifdir[entnum].marked?idx_marked:idx_black);
     /* box if cfg.xvpic_index and is an xvpic being used */
     if(cfg.xvpic_index && gifdir[entnum].xvw!=0)
       draw3dbox(xpos+(BARWIDTH-gifdir[entnum].xvw)/2-2,
-                ypos+GDFSIZ*6+39-gifdir[entnum].xvh/2-2,
+                ypos+GDFYBIT+39-gifdir[entnum].xvh/2-2,
                 xpos+(BARWIDTH-gifdir[entnum].xvw)/2+gifdir[entnum].xvw+1,
-                ypos+GDFSIZ*6+39-gifdir[entnum].xvh/2+gifdir[entnum].xvh+1,
+                ypos+GDFYBIT+39-gifdir[entnum].xvh/2+gifdir[entnum].xvh+1,
                 1,0, idx_light,idx_dark);
     }
   else
     {
     undraw3dbox(xpos,ypos,xpos+BARWIDTH-1,ypos+barheight-1,1);
-    undrawtext3d(xt,ypos+3,GDFSIZ,ctmp);
+    undrawtext3d(xt,ypos+3+gdfofs,gdfsiz,ctmp);
     vga_setcolor(gifdir[entnum].marked?idx_marked:idx_black);
-    vgadrawtext(xt,ypos+3,GDFSIZ,ctmp);
+    vgadrawtext(xt,ypos+3+gdfofs,gdfsiz,ctmp);
     /* undraw box if cfg.xvpic_index and is an xvpic being used */
     if(cfg.xvpic_index && gifdir[entnum].xvw!=0)
       undraw3dbox(xpos+(BARWIDTH-gifdir[entnum].xvw)/2-2,
-                  ypos+GDFSIZ*6+39-gifdir[entnum].xvh/2-2,
+                  ypos+GDFYBIT+39-gifdir[entnum].xvh/2-2,
                   xpos+(BARWIDTH-gifdir[entnum].xvw)/2+gifdir[entnum].xvw+1,
-                  ypos+GDFSIZ*6+39-gifdir[entnum].xvh/2+gifdir[entnum].xvh+1,
+                  ypos+GDFYBIT+39-gifdir[entnum].xvh/2+gifdir[entnum].xvh+1,
                   1);
     }
   }
 set_max_text_width(NO_CLIP_FONT);
-return;
 }
 
 
@@ -845,6 +930,7 @@ char cdir[MAXPATHLEN+1],*ptr;
 static char ctmp[1024],ctmp2[1024];
 int f,ypos,xpos,w,h,y,xt;
 unsigned char *image;
+int xvpics_dir_exists=0;
 
 getcwd(cdir,MAXPATHLEN);
 
@@ -862,6 +948,37 @@ if(drawdirmsg)
     drawtext3d(40,yofs+25,4,ctmp,1, idx_light,idx_dark,idx_black);
   }
 
+/* see if either xvpics dir exists. if not, we can speed things
+ * up a little by not bothering to look for thumbnails at all (the
+ * difference is usually small, but on a dos partition it's sometimes
+ * *really* painful otherwise).
+ *
+ * There's a little bit of duplicated code though, so it's a bit
+ * nasty and should be cleaned up sometime (XXX).
+ */
+if(strstr(cdir,"/.xvpics")!=NULL)
+  xvpics_dir_exists=1;
+else
+  {
+  struct stat sbuf;
+  
+  /* must be either a dir or symlink (symlink could be to a dir...) */
+  if(stat(".xvpics",&sbuf)!=-1 && (S_ISDIR(sbuf.st_mode) ||
+                                   S_ISLNK(sbuf.st_mode)))
+    xvpics_dir_exists=1;
+  else
+    {
+    sprintf(ctmp2,"%s/.xvpics/",getenv("HOME")?getenv("HOME"):"");
+    ptr=ctmp2+strlen(ctmp2);
+    getcwd(ptr,sizeof(ctmp2)-strlen(ctmp2)-1);
+    /* convert /foo/bar/baz to _foo_bar_baz */
+    while((ptr=strchr(ptr,'/'))!=NULL) *ptr++='_';
+    if(stat(ctmp2,&sbuf)!=-1 && (S_ISDIR(sbuf.st_mode) ||
+                                 S_ISLNK(sbuf.st_mode)))
+      xvpics_dir_exists=1;
+    }
+  }
+
 set_max_text_width(BAR_RESTRICT_WIDTH);
 for(f=startfrom;f<=gifdirsiz;f++)
   {
@@ -869,15 +986,15 @@ for(f=startfrom;f<=gifdirsiz;f++)
   if(xpos+BARWIDTH>COLS) break;
   ypos=fwinypos(f-startfrom+1);
   prettyfile(ctmp,&(gifdir[f]));
-  xt=cfg.xvpic_index?centreseltxt(xpos,GDFSIZ,ctmp):xpos+3;
+  xt=cfg.xvpic_index?centreseltxt(xpos,gdfsiz,ctmp):xpos+3;
   vga_setcolor(unshow?idx_medium:(gifdir[f].marked?idx_marked:idx_black));
-  vgadrawtext(xt,ypos+3,GDFSIZ,ctmp);
+  vgadrawtext(xt,ypos+3+gdfofs,gdfsiz,ctmp);
   if(cfg.thicktext && cfg.blockcursor)
-    vgadrawtext(xt+1,ypos+3,GDFSIZ,ctmp);
+    vgadrawtext(xt+1,ypos+3+gdfofs,gdfsiz,ctmp);
   vga_setcolor(unshow?idx_medium:idx_black);
   if(cfg.xvpic_index)
     {
-    /* load and draw index file (or undraw it) */
+    /* load and draw thumbnail file (or undraw it) */
     if(unshow)
       {
       image=malloc(96);
@@ -886,78 +1003,82 @@ for(f=startfrom;f<=gifdirsiz;f++)
         memset(image,idx_medium,96);
         for(y=-2;y<62;y++)
           vga_drawscansegment(image,
-          	xpos+(BARWIDTH-80)/2-2,ypos+y+GDFSIZ*6+9,96);
+          	xpos+(BARWIDTH-80)/2-2,ypos+y+GDFYBIT+9,96);
         free(image);
         }
       }
     else
       {
-      /* if '/.xvpics' is in the (absolute) path somewhere, we're
-       * in a xvpics subdir. So look for the thumbnail here! :-)
-       */
-      if(strstr(cdir,"/.xvpics")!=NULL)
-        strcpy(ctmp,gifdir[f].name);
-      else
-        sprintf(ctmp ,".xvpics/%s",gifdir[f].name);
-      sprintf(ctmp2,"%s/.xvpics/",getenv("HOME")?getenv("HOME"):"");
-      ptr=ctmp2+strlen(ctmp2);
-      getcwd(ptr,sizeof(ctmp2)-strlen(ctmp2)-1); ctmp2[strlen(ctmp2)]=0;
-      /* convert /foo/bar/baz to _foo_bar_baz */
-      while((ptr=strchr(ptr,'/'))!=NULL) *ptr++='_';
-      strcat(ctmp2,"/"); strcat(ctmp2,gifdir[f].name);
+      if(xvpics_dir_exists)
+        {
+        /* if '/.xvpics' is in the (absolute) path somewhere, we're
+         * in a xvpics subdir. So look for the thumbnail here! :-)
+         */
+        if(strstr(cdir,"/.xvpics")!=NULL)
+          strcpy(ctmp,gifdir[f].name);
+        else
+          sprintf(ctmp ,".xvpics/%s",gifdir[f].name);
+        sprintf(ctmp2,"%s/.xvpics/",getenv("HOME")?getenv("HOME"):"");
+        ptr=ctmp2+strlen(ctmp2);
+        getcwd(ptr,sizeof(ctmp2)-strlen(ctmp2)-1);
+        /* convert /foo/bar/baz to _foo_bar_baz */
+        while((ptr=strchr(ptr,'/'))!=NULL) *ptr++='_';
+        strcat(ctmp2,"/"); strcat(ctmp2,gifdir[f].name);
+        }
       
       gifdir[f].xvw=gifdir[f].xvh=0;
-      if(gifdir[f].isdir)
+      
+      if(xvpics_dir_exists && (read_xv332(ctmp,&image,&w,&h)==_PIC_OK ||
+         read_xv332(ctmp2,&image,&w,&h)==_PIC_OK))
         {
-        /* 'folder' icon, as usual for these type of things */
-        vga_setcolor(unshow?idx_medium:idx_black);
-        xt=xpos+(BARWIDTH-80)/2;
-        ypos+=GDFSIZ*6+9;
+        int xwant=xpos+(BARWIDTH-w)/2,w8=0;
         
-        /* main bit */
-        vga_drawline(xt+10,ypos+50,xt+70,ypos+50);
-        vga_drawline(xt+70,ypos+50,xt+70,ypos+20);
-        vga_drawline(xt+70,ypos+20,xt+65,ypos+15);
-        vga_drawline(xt+65,ypos+15,xt+15,ypos+15);
-        vga_drawline(xt+15,ypos+15,xt+10,ypos+20);
-        vga_drawline(xt+10,ypos+20,xt+10,ypos+50);
-        
-        /* top bit */
-        vga_drawline(xt+15,ypos+15,xt+20,ypos+10);
-        vga_drawline(xt+20,ypos+10,xt+35,ypos+10);
-        vga_drawline(xt+35,ypos+10,xt+40,ypos+15);
-        ypos-=GDFSIZ*6+9;
-        
-        gifdir[f].xvw=w=80; gifdir[f].xvh=h=60;
+        gifdir[f].xvw=w; gifdir[f].xvh=h;
+        if(v256==0)
+          {
+          w8=((w+7)&~7)+8;
+          greyfix332(&image,w,h,xwant&7,w8);
+          xwant&=~7;
+          }
+
+        if(image!=NULL)
+          {
+          for(y=0;y<h;y++)
+            vga_drawscansegment(image+y*(v256?w:w8),xwant,
+              ypos+y+GDFYBIT+39-h/2,v256?w:w8);
+          free(image);
+          }
         }
       else
-        if(read_xv332(ctmp ,&image,&w,&h)==_PIC_OK ||
-           read_xv332(ctmp2,&image,&w,&h)==_PIC_OK)
+        if(gifdir[f].isdir)
           {
-	  int xwant=xpos+(BARWIDTH-w)/2,w8=0;
-	  
-          gifdir[f].xvw=w; gifdir[f].xvh=h;
-	  if(v256==0)
-	    {
-	    w8=((w+7)&~7)+8;
-	    greyfix332(&image,w,h,xwant&7,w8);
-	    xwant&=~7;
-	    }
-
-          if(image!=NULL)
-            {
-            for(y=0;y<h;y++)
-              vga_drawscansegment(image+y*(v256?w:w8),xwant,
-                ypos+y+GDFSIZ*6+39-h/2,v256?w:w8);
-            free(image);
-            }
+          /* 'folder' icon, as usual for these types of things */
+          vga_setcolor(unshow?idx_medium:idx_black);
+          xt=xpos+(BARWIDTH-80)/2;
+          ypos+=GDFYBIT+9;
+          
+          /* main bit */
+          vga_drawline(xt+10,ypos+50,xt+70,ypos+50);
+          vga_drawline(xt+70,ypos+50,xt+70,ypos+20);
+          vga_drawline(xt+70,ypos+20,xt+65,ypos+15);
+          vga_drawline(xt+65,ypos+15,xt+15,ypos+15);
+          vga_drawline(xt+15,ypos+15,xt+10,ypos+20);
+          vga_drawline(xt+10,ypos+20,xt+10,ypos+50);
+          
+          /* top bit */
+          vga_drawline(xt+15,ypos+15,xt+20,ypos+10);
+          vga_drawline(xt+20,ypos+10,xt+35,ypos+10);
+          vga_drawline(xt+35,ypos+10,xt+40,ypos+15);
+          ypos-=GDFYBIT+9;
+          
+          gifdir[f].xvw=w=80; gifdir[f].xvh=h=60;
           }
         else
           {
           /* a default icon-type-thing here */
           vga_setcolor(unshow?idx_medium:idx_black);
           xt=xpos+(BARWIDTH-80)/2;
-          ypos+=GDFSIZ*6+9;
+          ypos+=GDFYBIT+9;
           
           /* main bit */
           vga_drawline(xt+20,ypos+50,xt+60,ypos+50);
@@ -970,14 +1091,14 @@ for(f=startfrom;f<=gifdirsiz;f++)
           vga_drawline(xt+50,ypos+10,xt+50,ypos+20);
           vga_drawline(xt+50,ypos+20,xt+60,ypos+20);
           
-          ypos-=GDFSIZ*6+9;
+          ypos-=GDFYBIT+9;
           gifdir[f].xvw=w=80; gifdir[f].xvh=h=60;
           }
       
       if(gifdir[f].xvw!=0)
         {
-        draw3dbox(xpos+(BARWIDTH-w)/2-1,ypos+GDFSIZ*6+38-h/2,
-                  xpos+(BARWIDTH-w)/2+w,ypos+GDFSIZ*6+39-h/2+h,1,0,
+        draw3dbox(xpos+(BARWIDTH-w)/2-1,ypos+GDFYBIT+38-h/2,
+                  xpos+(BARWIDTH-w)/2+w,ypos+GDFYBIT+39-h/2+h,1,0,
                   idx_light,idx_dark);
         }
       }
@@ -993,10 +1114,37 @@ DIR *dirfile;
 struct dirent *anentry;
 struct stat buf;
 char cdir[MAXPATHLEN+1];
-int entnum,l,isdir;
+int entnum,isdir;
 
 getcwd(cdir,MAXPATHLEN);
 dirfile=opendir(".");
+if(dirfile==NULL)
+  {
+  /* we get here if we can't read the dir.
+   * Zgv tests we have permission to access a file or dir before
+   * selecting it, so this can only happen if it was started on the dir
+   * from the cmdline, or if the directory has changed somehow since we
+   * last read it.
+   * the first reaction is to try $HOME instead.
+   * if *that* doesn't work, we cough and die, not unreasonably. :-)
+   */
+  /* be sure to mention what we're doing first... :-) */
+  msgbox(zgv_ttyfd,"Directory unreadable - going to $HOME...",
+  	MSGBOXTYPE_OK, idx_light,idx_dark,idx_black);
+  chdir(getenv("HOME"));
+  if((dirfile=opendir("."))==NULL)
+    {
+    screenoff();
+    fprintf(stderr,"$HOME is unreadable. This is a Bad Thing. TTFN...\n");
+    exit(1);
+    }
+  
+  /* this screenon() could result in us clearing the screen twice overall
+   * (if we did a ^R on a now-deleted dir, for example), but this isn't
+   * the kind of thing which happens all the time, so I guess this is ok.
+   */
+  screenon();
+  }
 entnum=0;
 while((anentry=readdir(dirfile))!=NULL)
   {
@@ -1012,29 +1160,13 @@ while((anentry=readdir(dirfile))!=NULL)
       buf.st_mode=0;
     isdir=S_ISDIR(buf.st_mode);
     /* directories, GIF/JPG/PBM/PGM/PPM tested here */
-    /* if I didn't know better I'd think this was a mess */
-    if(((l=strlen(anentry->d_name))>4) || isdir)
-      if((!strcasecmp(anentry->d_name+l-4,".gif")) ||
-         (!strcasecmp(anentry->d_name+l-4,".jpg")) ||
-         (!strcasecmp(anentry->d_name+l-5,".jpeg")) ||
-         (!strcasecmp(anentry->d_name+l-4,".bmp")) ||
-         (!strcasecmp(anentry->d_name+l-4,".tga")) ||
-#ifdef PNG_SUPPORT
-         (!strcasecmp(anentry->d_name+l-4,".png")) ||
-#endif
-#ifdef NGM_SUPPORT
-         (!strcasecmp(anentry->d_name+l-4,".ngm")) ||
-#endif
-         (!strcasecmp(anentry->d_name+l-4,".pcx")) ||
-         (!strcasecmp(anentry->d_name+l-4,".pbm")) ||
-         (!strcasecmp(anentry->d_name+l-4,".pgm")) ||
-         (!strcasecmp(anentry->d_name+l-4,".ppm")) || isdir)
-        {
-        entnum++;
-        gifdir[entnum].isdir=isdir;
-        strcpy(gifdir[entnum].name,anentry->d_name);
-        gifdir[entnum].marked=0;
-        }
+    if(isdir || ispicture(anentry->d_name))
+      {
+      entnum++;
+      gifdir[entnum].isdir=isdir;
+      strcpy(gifdir[entnum].name,anentry->d_name);
+      gifdir[entnum].marked=0;
+      }
     }
   }
 closedir(dirfile);
@@ -1043,10 +1175,36 @@ qsort(&(gifdir[1]),gifdirsiz,sizeof(struct GDIR),(void *)gcompare);
 }
 
 
+int ispicture(char *filename)
+{
+int l=strlen(filename);
+
+if(l<=4) return(0);
+
+if((!strcasecmp(filename+l-4,".gif")) ||
+   (!strcasecmp(filename+l-4,".jpg")) ||
+   (!strcasecmp(filename+l-5,".jpeg")) ||
+#ifdef PNG_SUPPORT
+   (!strcasecmp(filename+l-4,".png")) ||
+#endif
+   (!strcasecmp(filename+l-4,".mrf")) ||
+   (!strcasecmp(filename+l-4,".pbm")) ||
+   (!strcasecmp(filename+l-4,".pgm")) ||
+   (!strcasecmp(filename+l-4,".ppm")) ||
+   (!strcasecmp(filename+l-4,".bmp")) ||
+   (!strcasecmp(filename+l-4,".tga")) ||
+   (!strcasecmp(filename+l-4,".pcx")) ||
+   (!strcasecmp(filename+l-4,".ngm")))
+  return(1);
+else
+  return(0);
+}
+
+
 int gcompare(void *gn1,void *gn2)
 {
 struct GDIR *g1,*g2;
-char buf1[80],buf2[80];
+char buf1[258],buf2[258]; /* [sic] */
 
 g1=(struct GDIR *)gn1; g2=(struct GDIR *)gn2;
 prettyfile(buf1,g1); prettyfile(buf2,g2);
@@ -1072,6 +1230,10 @@ void screenon()
 int r,g,b,n;
 unsigned char *tmp;
 
+msgbox_draw_ok=1;
+gdfsiz=3-cfg.smallfiletext;
+gdfofs=6*cfg.smallfiletext;
+
 yofs=(cfg.fullsel?10:110);
 
 if(cfg.xvpic_index)
@@ -1090,11 +1252,11 @@ if(cfg.xvpic_index)
      * these are then blasted with the *real* file selector colours
      * unless 'perfectindex' is set.
      */
-    idx_medium=MAKE332COL(cfg.medium_r,cfg.medium_g,cfg.medium_b);
-    idx_dark  =MAKE332COL(cfg.dark_r  ,cfg.dark_g  ,cfg.dark_b  );
-    idx_light =MAKE332COL(cfg.light_r ,cfg.light_g ,cfg.light_b );
-    idx_black =MAKE332COL(cfg.black_r ,cfg.black_g ,cfg.black_b );
-    idx_marked=MAKE332COL(cfg.marked_r,cfg.marked_g,cfg.marked_b);
+    idx_medium=MAKE332COL(cfg.medium.r,cfg.medium.g,cfg.medium.b);
+    idx_dark  =MAKE332COL(cfg.dark.r  ,cfg.dark.g  ,cfg.dark.b  );
+    idx_light =MAKE332COL(cfg.light.r ,cfg.light.g ,cfg.light.b );
+    idx_black =MAKE332COL(cfg.black.r ,cfg.black.g ,cfg.black.b );
+    idx_marked=MAKE332COL(cfg.marked.r,cfg.marked.g,cfg.marked.b);
     }
   else
     {
@@ -1116,7 +1278,7 @@ if(cfg.xvpic_index)
     }
   
   /* fix bar height */
-  barheight=GDFSIZ*6+6+70;
+  barheight=GDFYBIT+6+70;
   }
 else
   {
@@ -1125,20 +1287,20 @@ else
    */
   vga_setmode((vga_hasmode(G640x480x256) && cfg.force16fs==0)?
                 G640x480x256:G640x480x16);
-  cfg.xvpic_index=0;
-  barheight=GDFSIZ*6+6;
+  barheight=GDFYBIT+6;
   idx_medium=MIDGREY; idx_dark=DARK; idx_light=LIGHT; idx_black=BLACK;
+  idx_marked=MARKEDCOL;
   if(one_file_only && vga_hasmode(G640x480x256) && cfg.force16fs==0)
     idx_medium=255;	/* well, not 0 at any rate */
   }
 
 if(!(cfg.xvpic_index && cfg.perfectindex))
   {
-  vga_setpalette(idx_medium,cfg.medium_r,cfg.medium_g,cfg.medium_b);
-  vga_setpalette(idx_dark  ,cfg.dark_r,cfg.dark_g,cfg.dark_b);
-  vga_setpalette(idx_light ,cfg.light_r,cfg.light_g,cfg.light_b);
-  vga_setpalette(idx_black ,cfg.black_r,cfg.black_g,cfg.black_b);
-  vga_setpalette(idx_marked,cfg.marked_r,cfg.marked_g,cfg.marked_b);
+  vga_setpalette(idx_medium,cfg.medium.r,cfg.medium.g,cfg.medium.b);
+  vga_setpalette(idx_dark  ,cfg.dark.r,cfg.dark.g,cfg.dark.b);
+  vga_setpalette(idx_light ,cfg.light.r,cfg.light.g,cfg.light.b);
+  vga_setpalette(idx_black ,cfg.black.r,cfg.black.g,cfg.black.b);
+  vga_setpalette(idx_marked,cfg.marked.r,cfg.marked.g,cfg.marked.b);
   }
 
 if(one_file_only) return;
@@ -1171,8 +1333,6 @@ else
 
 if(!cfg.fullsel)
   drawzgvlogo(10,10);
-
-return;
 }
 
 
@@ -1245,17 +1405,12 @@ switch(errnumber)
   case _PICERR_ISRLE:
     strcat(buf,"can't handle RLE BMP files"); break;
   case _PICERR_SHOWN_ALREADY:
-    /* this only happens with JPEGs */
+    /* this only happens with JPEGs/PNGs */
     return;
-#ifdef PNG_SUPPORT
-  case _PICERR_PNG_ERR:		/* a PNG error in zgv_pngerr */
-    strcpy(buf,zgv_pngerr); break;
-#endif /* PNG_SUPPORT */
   default:
     strcat(buf,"unknown error (ulp!)");
   }
 msgbox(zgv_ttyfd,buf,MSGBOXTYPE_OK, idx_light,idx_dark,idx_black);
-return;
 }
 
 
@@ -1307,7 +1462,7 @@ int done;
 done=sofar*100/total;
 
 vga_lockvc();
-if(vga_oktowrite() && ((done%5)==0 || sofar==total))
+if(vga_oktowrite() && ((done%10)==0 || sofar==total))
   {
   clear_xvpic(xv332_how_far_xpos,xv332_how_far_ypos);
 
@@ -1316,7 +1471,7 @@ if(vga_oktowrite() && ((done%5)==0 || sofar==total))
     vga_setcolor(idx_black);
     sprintf(tmp,"Reading - %2d%%",done);
     vgadrawtext(xv332_how_far_xpos+(BARWIDTH-70)/2,
-                xv332_how_far_ypos+GDFSIZ*6+39-4,2,tmp);
+                xv332_how_far_ypos+GDFYBIT+39-4,2,tmp);
     }
   }
 vga_unlockvc();
@@ -1330,7 +1485,7 @@ int y;
 
 memset(tmp,idx_medium,96);
 for(y=-3;y<63;y++)
-  vga_drawscansegment(tmp,xpos+(BARWIDTH-80)/2-3,ypos+y+GDFSIZ*6+9,96);
+  vga_drawscansegment(tmp,xpos+(BARWIDTH-80)/2-3,ypos+y+GDFYBIT+9,96);
 }
 
 
@@ -1349,7 +1504,7 @@ pixelsize=1;		/* ouch */
 if(howfar)
   {
   /* given the way the progress reporting from readpicture() works,
-   * I have to use some global variables here. :( 
+   * I have to use some global variables here. :-( 
    */
   xv332_how_far_xpos=howfar>>16;
   xv332_how_far_ypos=howfar&0xFFFF;
@@ -1366,7 +1521,7 @@ if(vga_oktowrite())
   {
   clear_xvpic(xv332_how_far_xpos,xv332_how_far_ypos);
   vgadrawtext(xv332_how_far_xpos+(BARWIDTH-62)/2,
-              xv332_how_far_ypos+GDFSIZ*6+39-4,2,"Resampling...");
+              xv332_how_far_ypos+GDFYBIT+39-4,2,"Resampling...");
   }
 vga_unlockvc();
 
@@ -1382,7 +1537,7 @@ if(vga_oktowrite())
   clear_xvpic(xv332_how_far_xpos,xv332_how_far_ypos);
   vga_setcolor(idx_black);
   vgadrawtext(xv332_how_far_xpos+(BARWIDTH-55)/2,
-              xv332_how_far_ypos+GDFSIZ*6+39-4,2,"Dithering...");
+              xv332_how_far_ypos+GDFYBIT+39-4,2,"Dithering...");
   }
 vga_unlockvc();
 
@@ -1411,27 +1566,171 @@ return(_PIC_OK);
 }
 
 
+/* hack of above routine to work on dirs.
+ * `filename' is the name of the dir in this case, of course.
+ */
+int makedirxv332(char *filename,char *xvpicfn,unsigned int howfar)
+{
+static char files[4][256];
+static struct { int x,y; } xypos[4]={{0,0}, {1,0}, {0,1}, {1,1}};
+FILE *out;
+DIR *dirfile;
+int tmp;
+int w,h,y;
+unsigned char *smallpic,*totalpic;
+struct dirent *anentry;
+struct stat buf;
+int f,pos;
+char tmpstr[40];
+
+/* the dir. xvpics are a montage of the top 4 files of subdir. working out
+ * which these are is even less fun than you'd think, as they aren't
+ * necessarily in alphabetical order. rather than read them all in, qsort
+ * that, and take the top 4, we instead keep a `top 4' table and do a kind
+ * of incremental sort on that as we go along.
+ */
+
+for(f=0;f<4;f++) files[f][0]=0;
+
+dirfile=opendir(filename);
+if(dirfile==NULL) return(_PICERR_NOFILE);
+
+while((anentry=readdir(dirfile))!=NULL)
+  {
+  if(anentry->d_name[0]=='.') continue;		/* skip `dot' files */
+  if((stat(anentry->d_name,&buf))==-1)
+    buf.st_mode=0;
+  if(S_ISDIR(buf.st_mode)) continue;		/* ...and subdirs */
+  if(!ispicture(anentry->d_name)) continue;	/* must be picture file */
+  
+  /* now check against existing files in table. start at the bottom
+   * and see how high we get.
+   */
+  pos=-1;
+  for(f=3;f>=0;f--)
+    if(files[f][0]==0 || strcmp(anentry->d_name,files[f])<0)
+      pos=f;
+  
+  if(pos>=0)
+    {
+    if(pos<3)
+      for(f=3;f>pos;f--)
+        strcpy(files[f],files[f-1]);
+    
+    strcpy(files[pos],anentry->d_name);
+    }
+  }
+closedir(dirfile);
+
+if(files[0][0]+files[1][0]+files[2][0]+files[3][0]==0)
+  /* no files there, just leave it alone. */
+  return(_PICERR_NOFILE);
+
+/* so we finally have the four filenames. */
+
+pixelsize=1;		/* ouch */
+
+if(howfar)
+  {
+  /* given the way the progress reporting from readpicture() works,
+   * I have to use some global variables here. :-( 
+   */
+  xv332_how_far_xpos=howfar>>16;
+  xv332_how_far_ypos=howfar&0xFFFF;
+  }
+
+/* get total image */
+if((totalpic=malloc(80*60))==NULL) return(_PICERR_NOMEM);
+
+memset(totalpic,0,80*60);
+
+for(f=0;f<4;f++)
+  {
+  vga_lockvc();
+  if(vga_oktowrite())
+    {
+    vga_setcolor(idx_black);
+    sprintf(tmpstr,"Image %d/4...",f+1);
+    clear_xvpic(xv332_how_far_xpos,xv332_how_far_ypos);
+    vgadrawtext(xv332_how_far_xpos+(BARWIDTH-62)/2,
+                xv332_how_far_ypos+GDFYBIT+39-4,2,tmpstr);
+    }
+  vga_unlockvc();
+  
+  tmp=((cfg.jpegindexstyle==3)?0:1);
+  sprintf(tmpstr,"%s/%s",filename,files[f]);
+  if((tmp=readpicture(tmpstr,NULL,0,tmp))!=_PIC_OK)
+    if(f==0)
+      {
+      free(totalpic);
+      remove(xvpicfn);
+      return(tmp);	/* if it's the first, abort... */
+      }
+    else
+      continue;		/* else just skip it. */
+  
+  /* image is pointed to by theimage, image_palette */
+  
+  /* resize */
+  w=40; h=30;
+  smallpic=resizepic(theimage,image_palette+512,image_palette+256,
+  		image_palette,width,height,&w,&h);
+  free(theimage); free(image_palette);	/* finished with these */
+  
+  /* dither */
+  ditherinit(w);
+  for(y=0;y<h;y++)
+    ditherline(smallpic+y*40*3,y,w);
+  ditherfinish();
+  
+  for(y=0;y<h;y++)
+    memcpy(totalpic+80*30*xypos[f].y+40*xypos[f].x+y*80,smallpic+y*40*3,w);
+  
+  free(smallpic);
+  }
+
+/* write */
+if((out=fopen(xvpicfn,"wb"))==NULL)
+  return(_PICERR_NOFILE);		/* well, kind of */
+
+fprintf(out,"P7 332\n");
+fprintf(out,"# dir thumbnail file produced by zgv v%s\n",ZGV_VER);
+fprintf(out,"#END_OF_COMMENTS\n");
+fprintf(out,"80 60 255\n");
+
+fwrite(totalpic,1,80*60,out);
+
+fclose(out);
+
+free(totalpic);
+
+return(_PIC_OK);
+}
+
+
 /* update indexes (xvpics) of current directory.
  * we draw the stuff as we go along, and allow Esc to abort
  * the process. Checking for Esc happens between creating xvpics.
  * the directory '.xvpics' is created if needed.
  */
-void update_xvpics()
+void update_xvpics(int do_dirs_instead)
 {
 FILE *test;
 int f;
 int curent,oldstart,oldent,startfrom;
-struct stat realpic,xvpic,xvpic2;
-char buf[1024],buf2[1024],*ptr;
+struct stat realpic,xvpic,xvpic2,tmpsbuf;
+static char buf[1024],buf2[1024],altdirbuf[1024];
+char *ptr;
 int redraw_needed=0;
 int r1=0,r2=0;
+int altdirexists=0;
 
 
 /* for each picture in the current directory, we check to see if
  * a file exists in the .xvpics directory with the same filename.
  * If not, it is created. If a file does exist, we check the
  * modification times. If the picture file is newer than the index
- * file, a new one gets created. Hope that's clear now. :)
+ * file, a new one gets created. Hope that's clear now. :-)
  */
 
 redrawall(1,1);
@@ -1439,28 +1738,48 @@ curent=startfrom=1;
 
 vga_runinbackground(1);
 
+
+/* see if the ~/.xvpics/foo_bar_baz dir exists. If not, we can
+ * avoid having to test for it below, making for far fewer stat()s.
+ * this also sets altdirbuf, so we don't have to work out the dir's
+ * name all the time. :-)
+ */
+
+sprintf(altdirbuf,"%s/.xvpics/",getenv("HOME")?getenv("HOME"):"");
+ptr=altdirbuf+strlen(altdirbuf);
+getcwd(ptr,sizeof(altdirbuf)-strlen(altdirbuf)-1);
+while((ptr=strchr(ptr,'/'))!=NULL) *ptr++='_';
+altdirexists=(stat(altdirbuf,&tmpsbuf)!=-1);
+
+
 for(f=1;f<=gifdirsiz;f++)
   {
-  if(gifdir[f].isdir==0)
+  /* with the do_dirs_instead mode, the first non-dir entry is a
+   * sensible place to stop.
+   */
+  if(do_dirs_instead && gifdir[f].isdir==0) break;
+  
+  if(gifdir[f].isdir==do_dirs_instead && stat(gifdir[f].name,&realpic)!=-1 &&
+  	gifdir[f].name[0]!='.')
     {
-    if(stat(gifdir[f].name,&realpic)==-1)
-      continue;			/* the picture file doesn't exist */
-    
     /* test for normal .xvpics/wibble */
     sprintf(buf,".xvpics/%s",gifdir[f].name);
+    r1=stat(buf,&xvpic);
+    
+    /* and for ~/.xvpics/foo_bar_baz/wibble if needed */
+    if(!altdirexists)
+      r2=-1;
+    else
+      {
+      strcpy(buf2,altdirbuf);
+      strcat(buf2,"/"); strcat(buf2,gifdir[f].name);
+      r2=stat(buf2,&xvpic2);
+      }
     
-    /* and for ~/.xvpics/foo_bar_baz/wibble */
-    sprintf(buf2,"%s/.xvpics/",getenv("HOME")?getenv("HOME"):"");
-    ptr=buf2+strlen(buf2);
-    getcwd(ptr,sizeof(buf2)-strlen(buf2)-1); buf2[strlen(buf2)]=0;
-    while((ptr=strchr(ptr,'/'))!=NULL) *ptr++='_';
-    strcat(buf2,"/"); strcat(buf2,gifdir[f].name);
-
-    r1=stat(buf,&xvpic); r2=stat(buf2,&xvpic2);
-            
     if((r1==-1 && r2==-1) ||
        (r1!=-1 && realpic.st_mtime>xvpic.st_mtime) ||
-       (r2!=-1 && realpic.st_mtime>xvpic2.st_mtime))
+       (r2!=-1 && realpic.st_mtime>xvpic2.st_mtime) ||
+       do_dirs_instead)
       {
       /* this is pretty BFI and messy */
       
@@ -1486,26 +1805,25 @@ for(f=1;f<=gifdirsiz;f++)
           }
         
         /* also need to create ~/.xvpics/_foo_bar_baz */
-        strcat(buf,"/");
-        ptr=buf+strlen(buf);
-        getcwd(ptr,sizeof(buf)-strlen(buf)-1); buf[strlen(buf)]=0;
-        /* convert /foo/bar/baz to _foo_bar_baz */
-        while((ptr=strchr(ptr,'/'))!=NULL) *ptr++='_';
-        
-        if(mkdir(buf,0777)==-1 && errno!=EEXIST && errno!=EACCES)
+        if(mkdir(altdirbuf,0777)==-1 && errno!=EEXIST && errno!=EACCES)
           {
           wait_for_foreground();
-          msgbox(zgv_ttyfd,"Unable to create -/.xvpics/... directory",
+          msgbox(zgv_ttyfd,"Unable to create ~/.xvpics/... directory",
           		MSGBOXTYPE_OK,idx_light,idx_dark,idx_black);
           return;
           }
         
         /* make sure filename for xvpic is in buf */
-        strcat(buf,"/"); strcat(buf,gifdir[f].name);
+        strcpy(buf,altdirbuf); strcat(buf,"/"); strcat(buf,gifdir[f].name);
         }
       
-      makexv332(gifdir[f].name,buf,(fwinxpos(f-startfrom+1)<<16) |
-      					fwinypos(f-startfrom+1));
+      if(do_dirs_instead)
+        makedirxv332(gifdir[f].name,buf,(fwinxpos(f-startfrom+1)<<16) |
+        					fwinypos(f-startfrom+1));
+      else
+        makexv332(gifdir[f].name,buf,(fwinxpos(f-startfrom+1)<<16) |
+        					fwinypos(f-startfrom+1));
+      
       /* redraw without mode change */
       vga_lockvc();
       if(vga_oktowrite())
@@ -1525,7 +1843,8 @@ for(f=1;f<=gifdirsiz;f++)
       vga_unlockvc();
       }
     }
-    
+  
+  
   /* move down one if possible */
   oldent=curent; oldstart=startfrom;
   if(curent<gifdirsiz) curent++;
@@ -1563,15 +1882,18 @@ for(f=1;f<=gifdirsiz;f++)
   }
 
 wait_for_foreground();
-usleep(400000);	/* for effect :) */
-return;
+usleep(400000);	/* for effect :-) */
 }
 
 
-
 /* if we're not running directly on a console, try to find a free console
  * and move us to it. Notes old VT so we can switch back to it when finished,
  * if we ran on a different one to start off with.
+ *
+ * svgalib 1.2.11 and up do something similar (which was based on this,
+ * I think), but it doesn't *quite* do what we need (e.g. it changes stdout,
+ * which is sensible generally but zgv wants to keep it the same). So, er,
+ * that's why this code's still here. :-)
  */
 int fixvt()
 {
@@ -1775,7 +2097,6 @@ else
     
 free(*image);
 *image=new;
-return;
 }
 
 
