@@ -1,10 +1,10 @@
-/* Zgv v2.8 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
- * Copyright (C) 1993-1997 Russell Marks. See README for license details.
+/* Zgv v3.0 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993-1998 Russell Marks. See README for license details.
  *
  * zgv.h - constants, typedefs.
  */
 
-#define ZGV_VER		"2.8"
+#define ZGV_VER		"3.0"
 
 
 #define _PIC_OK			 0
@@ -17,7 +17,6 @@
 #define _PICERR_CORRUPT		-7
 #define _PICERR_SHOWN_ALREADY	-8
 #define _PICERR_ISRLE		-9
-#define _PICERR_PNG_ERR		-10
 
 typedef struct {
   int width,height;
@@ -32,6 +31,9 @@ typedef unsigned char byte;
 extern int idx_light,idx_medium,idx_dark,idx_black;
 extern int tagview_mode;
 extern int zgv_ttyfd;
-extern char zgv_pngerr[];
 
 extern void wait_for_foreground(void);
+
+/* vgadrawtext_bmp() needs to know what the current colour is, so... */
+extern int current_colour;
+#define vga_setcolor(x) vga_setcolor(current_colour=(x))
